# -*- coding: utf-8 -*-
"""
    emmett.security
    ---------------

    Miscellaneous security helpers.

    :copyright: 2014 Giovanni Barillari

    Based on the code of web2py (http://www.web2py.com)
    :copyright: (c) by Massimo Di Pierro <mdipierro@cs.depaul.edu>

    :license: LGPLv3 (http://www.gnu.org/licenses/lgpl.html)
"""

import base64
import hashlib
import hmac
import os
import pickle
import pyaes
import random
import struct
import threading
import time
import uuid as uuidm
import zlib

from collections import OrderedDict

# TODO: check bytes conversions
from ._shortcuts import hashlib_sha1, to_bytes
from .libs.pbkdf2 import pbkdf2_hex

try:
    from emmett_crypto import kdf
except:
    kdf = None


class CSRFStorage(OrderedDict):
    def _clean(self):
        now = time.time()
        for key in list(self):
            if self[key] + 3600 > now:
                break
            del self[key]

    def gen_token(self):
        self._clean()
        token = str(uuid())
        self[token] = int(time.time())
        return token


def md5_hash(text):
    """ Generate a md5 hash with the given text """
    return hashlib.md5(text).hexdigest()


def simple_hash(text, key='', salt='', digest_alg='md5'):
    """
    Generates hash with the given text using the specified
    digest hashing algorithm
    """
    if not digest_alg:
        raise RuntimeError("simple_hash with digest_alg=None")
    elif not isinstance(digest_alg, str):  # manual approach
        h = digest_alg(text + key + salt)
    elif digest_alg.startswith('pbkdf2'):  # latest and coolest!
        iterations, keylen, alg = digest_alg[7:-1].split(',')
        if kdf:
            return kdf.pbkdf2_hex(
                text,
                salt,
                iterations=int(iterations),
                keylen=int(keylen),
                hash_algorithm=kdf.PBKDF2_HMAC[alg]
            )
        return pbkdf2_hex(
            to_bytes(text),
            to_bytes(salt),
            int(iterations),
            int(keylen),
            get_digest(alg)
        )
    elif key:  # use hmac
        digest_alg = get_digest(digest_alg)
        h = hmac.new(to_bytes(key + salt), msg=to_bytes(text), digestmod=digest_alg)
    else:  # compatible with third party systems
        h = hashlib.new(digest_alg)
        h.update(to_bytes(text + salt))
    return h.hexdigest()


def get_digest(value):
    """
    Returns a hashlib digest algorithm from a string
    """
    if not isinstance(value, str):
        return value
    value = value.lower()
    if value == "md5":
        return hashlib.md5
    elif value == "sha1":
        return hashlib.sha1
    elif value == "sha224":
        return hashlib.sha224
    elif value == "sha256":
        return hashlib.sha256
    elif value == "sha384":
        return hashlib.sha384
    elif value == "sha512":
        return hashlib.sha512
    else:
        raise ValueError("Invalid digest algorithm: %s" % value)


DIGEST_ALG_BY_SIZE = {
    128 / 4: 'md5',
    160 / 4: 'sha1',
    224 / 4: 'sha224',
    256 / 4: 'sha256',
    384 / 4: 'sha384',
    512 / 4: 'sha512',
}


def _pad(s, n=32, padchar='.'):
    expected_len = ((len(s) + n) - len(s) % n)
    return s.ljust(expected_len, to_bytes(padchar))


# DEPRECATED: remove this method in future versions
def secure_dumps(data, encryption_key, hash_key=None, compression_level=None):
    if not hash_key:
        hash_key = hashlib_sha1(encryption_key).hexdigest()
    dump = pickle.dumps(data)
    if compression_level:
        dump = zlib.compress(dump, compression_level)
    key = _pad(to_bytes(encryption_key[:32]))
    aes = pyaes.AESModeOfOperationCFB(key, iv=key[:16], segment_size=8)
    encrypted_data = base64.urlsafe_b64encode(aes.encrypt(_pad(dump)))
    signature = hmac.new(to_bytes(hash_key), msg=encrypted_data, digestmod='md5').hexdigest()
    return signature + ':' + encrypted_data.decode('utf8')


# DEPRECATED: remove this method in future versions
def secure_loads(data, encryption_key, hash_key=None, compression_level=None):
    if ':' not in data:
        return None
    if not hash_key:
        hash_key = hashlib_sha1(encryption_key).hexdigest()
    signature, encrypted_data = data.split(':', 1)
    actual_signature = hmac.new(
        to_bytes(hash_key), msg=to_bytes(encrypted_data), digestmod='md5').hexdigest()
    if signature != actual_signature:
        return None
    key = _pad(to_bytes(encryption_key[:32]))
    aes = pyaes.AESModeOfOperationCFB(key, iv=key[:16], segment_size=8)
    try:
        data = aes.decrypt(base64.urlsafe_b64decode(to_bytes(encrypted_data)))
        data = data.rstrip(to_bytes(' '))
        if compression_level:
            data = zlib.decompress(data)
        return pickle.loads(data)
    except (TypeError, pickle.UnpicklingError):
        return None


def _init_urandom():
    """
    This function and the web2py_uuid follow from the following discussion:
    http://groups.google.com/group/web2py-developers/browse_thread/thread/7fd5789a7da3f09

    At startup web2py compute a unique ID that identifies the machine by adding
    uuid.getnode() + int(time.time() * 1e3)

    This is a 48-bit number. It converts the number into 16 8-bit tokens.
    It uses this value to initialize the entropy source ('/dev/urandom')
    and to seed random.

    If os.random() is not supported, it falls back to using random and issues
    a warning.
    """
    node_id = uuidm.getnode()
    microseconds = int(time.time() * 1e6)
    ctokens = [((node_id + microseconds) >> ((i % 6) * 8)) %
               256 for i in range(16)]
    random.seed(node_id + microseconds)
    try:
        os.urandom(1)
        have_urandom = True
        try:
            # try to add process-specific entropy
            frandom = open('/dev/urandom', 'wb')
            try:
                frandom.write(bytes([]).join(bytes([t]) for t in ctokens))
            finally:
                frandom.close()
        except IOError:
            # works anyway
            pass
    except NotImplementedError:
        have_urandom = False
    packed = bytes([]).join(bytes([x]) for x in ctokens)
    unpacked_ctokens = struct.unpack('=QQ', packed)
    return unpacked_ctokens, have_urandom


_UNPACKED_CTOKENS, _HAVE_URANDOM = _init_urandom()


def fast_urandom16(urandom=[], locker=threading.RLock()):
    """
    this is 4x faster than calling os.urandom(16) and prevents
    the "too many files open" issue with concurrent access to os.urandom()
    """
    try:
        return urandom.pop()
    except IndexError:
        try:
            locker.acquire()
            ur = os.urandom(16 * 1024)
            urandom += [ur[i:i + 16] for i in range(16, 1024 * 16, 16)]
            return ur[0:16]
        finally:
            locker.release()


def uuid(ctokens=_UNPACKED_CTOKENS):
    """
    It works like uuid.uuid4 except that tries to use os.urandom() if possible
    and it XORs the output with the tokens uniquely associated with
    this machine.
    """
    rand_longs = (random.getrandbits(64), random.getrandbits(64))
    if _HAVE_URANDOM:
        urand_longs = struct.unpack('=QQ', fast_urandom16())
        byte_s = struct.pack('=QQ',
                             rand_longs[0] ^ urand_longs[0] ^ ctokens[0],
                             rand_longs[1] ^ urand_longs[1] ^ ctokens[1])
    else:
        byte_s = struct.pack('=QQ',
                             rand_longs[0] ^ ctokens[0],
                             rand_longs[1] ^ ctokens[1])
    return str(uuidm.UUID(bytes=byte_s, version=4))



# -*- coding: utf-8 -*-
"""
    emmett.helpers
    --------------

    Provides helping methods for applications.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import os
import re

from typing import Any, List, Optional, Tuple, Union

from pydal.exceptions import NotAuthorizedException, NotFoundException

from .ctx import current
from .html import HtmlTag, tag
from .http import HTTP, HTTPFile, HTTPIO

_re_dbstream = re.compile(r'(?P<table>.*?)\.(?P<field>.*?)\..*')


def abort(code: int, body: str = ''):
    response = current.response
    response.status = code
    raise HTTP(
        code,
        body=body,
        cookies=response.cookies
    )


def stream_file(path: str):
    full_path = os.path.join(current.app.root_path, path)
    raise HTTPFile(
        full_path,
        headers=current.response.headers,
        cookies=current.response.cookies
    )


def stream_dbfile(db: Any, name: str):
    items = _re_dbstream.match(name)
    if not items:
        abort(404)
    table_name, field_name = items.group('table'), items.group('field')
    try:
        field = db[table_name][field_name]
    except AttributeError:
        abort(404)
    try:
        filename, path_or_stream = field.retrieve(name, nameonly=True)
    except NotAuthorizedException:
        abort(403)
    except NotFoundException:
        abort(404)
    except IOError:
        abort(404)
    if isinstance(path_or_stream, str):
        raise HTTPFile(
            path_or_stream,
            headers=current.response.headers,
            cookies=current.response.cookies
        )
    raise HTTPIO(
        path_or_stream,
        headers=current.response.headers,
        cookies=current.response.cookies
    )


def flash(message: str, category: str = 'message'):
    #: Flashes a message to the next request.
    if current.session._flashes is None:
        current.session._flashes = []
    current.session._flashes.append((category, message))


def get_flashed_messages(
    with_categories: bool = False,
    category_filter: Union[str, List[str]] = []
) -> Union[List[str], Tuple[str, str]]:
    #: Pulls flashed messages from the session and returns them.
    #  By default just the messages are returned, but when `with_categories`
    #  is set to `True`, the return value will be a list of tuples in the
    #  form `(category, message)` instead.
    if not isinstance(category_filter, list):
        category_filter = [category_filter]
    try:
        flashes = list(current.session._flashes or [])
        if category_filter:
            flashes = list(filter(lambda f: f[0] in category_filter, flashes))
        for el in flashes:
            current.session._flashes.remove(el)
        if not with_categories:
            return [x[1] for x in flashes]
    except Exception:
        flashes = []
    return flashes


def load_component(
    url: str,
    target: Optional[str] = None,
    content: str = 'loading...'
) -> HtmlTag:
    attr = {}
    if target:
        attr['_id'] = target
    attr['_data-emt_remote'] = url
    return tag.div(content, **attr)



# -*- coding: utf-8 -*-
"""
    emmett.validators.helpers
    -------------------------

    Provides utilities for validators.

    Ported from the original validators of web2py (http://www.web2py.com)

    :copyright: (c) by Massimo Di Pierro <mdipierro@cs.depaul.edu>
    :license: LGPLv3 (http://www.gnu.org/licenses/lgpl.html)
"""

import re

from datetime import tzinfo, timedelta
from io import StringIO

# TODO: check unicode conversions
from .._shortcuts import to_unicode
from ..ctx import current
from ..security import simple_hash, uuid, DIGEST_ALG_BY_SIZE


_DEFAULT = lambda: None


def translate(text):
    if text is None:
        return None
    elif isinstance(text, str):
        return current.T(text)
    return str(text)


def options_sorter(x, y):
    return (str(x[1]).upper() > str(y[1]).upper() and 1) or -1


def is_empty(value, empty_regex=None):
    if value is None:
        return value, True
    if isinstance(value, (str, bytes)):
        value = value.strip()
        if empty_regex is not None and empty_regex.match(value):
            value = ''
        return value, len(value) == 0
    if isinstance(value, list):
        return value, len(value) == 0
    return value, False


class _UTC(tzinfo):
    ZERO = timedelta(0)

    def utcoffset(self, dt):
        return _UTC.ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return _UTC.ZERO


class LazyCrypt(object):
    def __init__(self, crypt, password):
        """
        crypt is an instance of the Crypt validator,
        password is the password as inserted by the user
        """
        self.crypt = crypt
        self.password = password
        self.crypted = None

    def __str__(self):
        """
        Encrypted self.password and caches it in self.crypted.
        If self.crypt.salt the output is in the format
        <algorithm>$<salt>$<hash>

        Try get the digest_alg from the key (if it exists)
        else assume the default digest_alg. If not key at all, set key=''

        If a salt is specified use it, if salt is True, set salt to uuid
        (this should all be backward compatible)

        Options:
        key = 'uuid'
        key = 'md5:uuid'
        key = 'sha512:uuid'
        ...
        key = 'pbkdf2(1000,64,sha512):uuid' 1000 iterations and 64 chars length
        """
        if self.crypted:
            return self.crypted
        if self.crypt.key:
            if ':' in self.crypt.key:
                digest_alg, key = self.crypt.key.split(':', 1)
            else:
                digest_alg, key = self.crypt.digest_alg, self.crypt.key
        else:
            digest_alg, key = self.crypt.digest_alg, ''
        if self.crypt.salt:
            if self.crypt.salt == True:
                salt = str(uuid()).replace('-', '')[-16:]
            else:
                salt = self.crypt.salt
        else:
            salt = ''
        hashed = simple_hash(self.password, key, salt, digest_alg)
        self.crypted = '%s$%s$%s' % (digest_alg, salt, hashed)
        return self.crypted

    def __eq__(self, stored_password):
        """
        compares the current lazy crypted password with a stored password
        """
        if isinstance(stored_password, self.__class__):
            return ((self is stored_password) or
                    ((self.crypt.key == stored_password.crypt.key) and
                    (self.password == stored_password.password)))

        if self.crypt.key:
            if ':' in self.crypt.key:
                key = self.crypt.key.split(':')[1]
            else:
                key = self.crypt.key
        else:
            key = ''
        if stored_password is None:
            return False
        elif stored_password.count('$') == 2:
            (digest_alg, salt, hash) = stored_password.split('$')
            h = simple_hash(self.password, key, salt, digest_alg)
            temp_pass = '%s$%s$%s' % (digest_alg, salt, h)
        else:  # no salting
            # guess digest_alg
            digest_alg = DIGEST_ALG_BY_SIZE.get(len(stored_password), None)
            if not digest_alg:
                return False
            else:
                temp_pass = simple_hash(self.password, key, '', digest_alg)
        return temp_pass == stored_password

    def __ne__(self, other):
        return not self.__eq__(other)


def _escape_unicode(string):
    '''
    Converts a unicode string into US-ASCII, using a simple conversion scheme.
    Each unicode character that does not have a US-ASCII equivalent is
    converted into a URL escaped form based on its hexadecimal value.
    For example, the unicode character '\u4e86' will become the string '%4e%86'

    Args:
        string: unicode string, the unicode string to convert into an
            escaped US-ASCII form

    Returns:
        string: the US-ASCII escaped form of the inputted string

    @author: Jonathan Benn
    '''
    returnValue = StringIO()

    for character in string:
        code = ord(character)
        if code > 0x7F:
            hexCode = hex(code)
            returnValue.write('%' + hexCode[2:4] + '%' + hexCode[4:6])
        else:
            returnValue.write(character)

    return returnValue.getvalue()


def _unicode_to_ascii_authority(authority):
    '''
    Follows the steps in RFC 3490, Section 4 to convert a unicode authority
    string into its ASCII equivalent.
    For example, u'www.Alliancefran\xe7aise.nu' will be converted into
    'www.xn--alliancefranaise-npb.nu'

    Args:
        authority: unicode string, the URL authority component to convert,
            e.g. u'www.Alliancefran\xe7aise.nu'

    Returns:
        string: the US-ASCII character equivalent to the inputed authority,
             e.g. 'www.xn--alliancefranaise-npb.nu'

    Raises:
        Exception: if the function is not able to convert the inputed
            authority

    @author: Jonathan Benn
    '''
    label_split_regex = re.compile(u'[\u002e\u3002\uff0e\uff61]')

    #RFC 3490, Section 4, Step 1
    #The encodings.idna Python module assumes that AllowUnassigned == True

    #RFC 3490, Section 4, Step 2
    labels = label_split_regex.split(authority)

    #RFC 3490, Section 4, Step 3
    #The encodings.idna Python module assumes that UseSTD3ASCIIRules == False

    #RFC 3490, Section 4, Step 4
    #We use the ToASCII operation because we are about to put the authority
    #into an IDN-unaware slot
    asciiLabels = []
    try:
        import encodings.idna
        for label in labels:
            if label:
                asciiLabels.append(encodings.idna.ToASCII(label))
            else:
                #encodings.idna.ToASCII does not accept an empty string, but
                #it is necessary for us to allow for empty labels so that we
                #don't modify the URL
                asciiLabels.append('')
    except:
        asciiLabels = [str(label) for label in labels]
    #RFC 3490, Section 4, Step 5
    return str(reduce(lambda x, y: x + unichr(0x002E) + y, asciiLabels))


def unicode_to_ascii_url(url, prepend_scheme):
    '''
    Converts the inputed unicode url into a US-ASCII equivalent. This function
    goes a little beyond RFC 3490, which is limited in scope to the domain name
    (authority) only. Here, the functionality is expanded to what was observed
    on Wikipedia on 2009-Jan-22:

       Component    Can Use Unicode?
       ---------    ----------------
       scheme       No
       authority    Yes
       path         Yes
       query        Yes
       fragment     No

    The authority component gets converted to punycode, but occurrences of
    unicode in other components get converted into a pair of URI escapes (we
    assume 4-byte unicode). E.g. the unicode character U+4E2D will be
    converted into '%4E%2D'. Testing with Firefox v3.0.5 has shown that it can
    understand this kind of URI encoding.

    Args:
        url: unicode string, the URL to convert from unicode into US-ASCII
        prepend_scheme: string, a protocol scheme to prepend to the URL if
            we're having trouble parsing it.
            e.g. "http". Input None to disable this functionality

    Returns:
        string: a US-ASCII equivalent of the inputed url

    @author: Jonathan Benn
    '''
    #convert the authority component of the URL into an ASCII punycode string,
    #but encode the rest using the regular URI character encoding

    groups = url_split_regex.match(url).groups()
    #If no authority was found
    if not groups[3]:
        #Try appending a scheme to see if that fixes the problem
        scheme_to_prepend = prepend_scheme or 'http'
        groups = url_split_regex.match(
            to_unicode(scheme_to_prepend) + u'://' + url).groups()
    #if we still can't find the authority
    if not groups[3]:
        raise Exception('No authority component found, ' +
                        'could not decode unicode to US-ASCII')

    #We're here if we found an authority, let's rebuild the URL
    scheme = groups[1]
    authority = groups[3]
    path = groups[4] or ''
    query = groups[5] or ''
    fragment = groups[7] or ''

    if prepend_scheme:
        scheme = str(scheme) + '://'
    else:
        scheme = ''
    return scheme + _unicode_to_ascii_authority(authority) +\
        _escape_unicode(path) + _escape_unicode(query) + str(fragment)


url_split_regex = \
    re.compile(r'^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?')

official_url_schemes = [
    'aaa', 'aaas', 'acap', 'cap', 'cid', 'crid', 'data', 'dav', 'dict',
    'dns', 'fax', 'file', 'ftp', 'go', 'gopher', 'h323', 'http', 'https',
    'icap', 'im', 'imap', 'info', 'ipp', 'iris', 'iris.beep', 'iris.xpc',
    'iris.xpcs', 'iris.lws', 'ldap', 'mailto', 'mid', 'modem', 'msrp',
    'msrps', 'mtqp', 'mupdate', 'news', 'nfs', 'nntp', 'opaquelocktoken',
    'pop', 'pres', 'prospero', 'rtsp', 'service', 'shttp', 'sip', 'sips',
    'snmp', 'soap.beep', 'soap.beeps', 'tag', 'tel', 'telnet', 'tftp',
    'thismessage', 'tip', 'tv', 'urn', 'vemmi', 'wais', 'xmlrpc.beep',
    'xmlrpc.beep', 'xmpp', 'z39.50r', 'z39.50s'
]

unofficial_url_schemes = [
    'about', 'adiumxtra', 'aim', 'afp', 'aw', 'callto', 'chrome', 'cvs',
    'ed2k', 'feed', 'fish', 'gg', 'gizmoproject', 'iax2', 'irc', 'ircs',
    'itms', 'jar', 'javascript', 'keyparc', 'lastfm', 'ldaps', 'magnet',
    'mms', 'msnim', 'mvn', 'notes', 'nsfw', 'psyc', 'paparazzi:http',
    'rmi', 'rsync', 'secondlife', 'sgn', 'skype', 'ssh', 'sftp', 'smb',
    'sms', 'soldat', 'steam', 'svn', 'teamspeak', 'unreal', 'ut2004',
    'ventrilo', 'view-source', 'webcal', 'wyciwyg', 'xfire', 'xri', 'ymsgr'
]

official_top_level_domains = [
    # a
    'abogado', 'ac', 'academy', 'accountants', 'active', 'actor',
    'ad', 'adult', 'ae', 'aero', 'af', 'ag', 'agency', 'ai',
    'airforce', 'al', 'allfinanz', 'alsace', 'am', 'amsterdam', 'an',
    'android', 'ao', 'apartments', 'aq', 'aquarelle', 'ar', 'archi',
    'army', 'arpa', 'as', 'asia', 'associates', 'at', 'attorney',
    'au', 'auction', 'audio', 'autos', 'aw', 'ax', 'axa', 'az',
    # b
    'ba', 'band', 'bank', 'bar', 'barclaycard', 'barclays',
    'bargains', 'bayern', 'bb', 'bd', 'be', 'beer', 'berlin', 'best',
    'bf', 'bg', 'bh', 'bi', 'bid', 'bike', 'bingo', 'bio', 'biz',
    'bj', 'black', 'blackfriday', 'bloomberg', 'blue', 'bm', 'bmw',
    'bn', 'bnpparibas', 'bo', 'boo', 'boutique', 'br', 'brussels',
    'bs', 'bt', 'budapest', 'build', 'builders', 'business', 'buzz',
    'bv', 'bw', 'by', 'bz', 'bzh',
    # c
    'ca', 'cab', 'cal', 'camera', 'camp', 'cancerresearch', 'canon',
    'capetown', 'capital', 'caravan', 'cards', 'care', 'career',
    'careers', 'cartier', 'casa', 'cash', 'casino', 'cat',
    'catering', 'cbn', 'cc', 'cd', 'center', 'ceo', 'cern', 'cf',
    'cg', 'ch', 'channel', 'chat', 'cheap', 'christmas', 'chrome',
    'church', 'ci', 'citic', 'city', 'ck', 'cl', 'claims',
    'cleaning', 'click', 'clinic', 'clothing', 'club', 'cm', 'cn',
    'co', 'coach', 'codes', 'coffee', 'college', 'cologne', 'com',
    'community', 'company', 'computer', 'condos', 'construction',
    'consulting', 'contractors', 'cooking', 'cool', 'coop',
    'country', 'cr', 'credit', 'creditcard', 'cricket', 'crs',
    'cruises', 'cu', 'cuisinella', 'cv', 'cw', 'cx', 'cy', 'cymru',
    'cz',
    # d
    'dabur', 'dad', 'dance', 'dating', 'day', 'dclk', 'de', 'deals',
    'degree', 'delivery', 'democrat', 'dental', 'dentist', 'desi',
    'design', 'dev', 'diamonds', 'diet', 'digital', 'direct',
    'directory', 'discount', 'dj', 'dk', 'dm', 'dnp', 'do', 'docs',
    'domains', 'doosan', 'durban', 'dvag', 'dz',
    # e
    'eat', 'ec', 'edu', 'education', 'ee', 'eg', 'email', 'emerck',
    'energy', 'engineer', 'engineering', 'enterprises', 'equipment',
    'er', 'es', 'esq', 'estate', 'et', 'eu', 'eurovision', 'eus',
    'events', 'everbank', 'exchange', 'expert', 'exposed',
    # f
    'fail', 'fans', 'farm', 'fashion', 'feedback', 'fi', 'finance',
    'financial', 'firmdale', 'fish', 'fishing', 'fit', 'fitness',
    'fj', 'fk', 'flights', 'florist', 'flowers', 'flsmidth', 'fly',
    'fm', 'fo', 'foo', 'football', 'forsale', 'foundation', 'fr',
    'frl', 'frogans', 'fund', 'furniture', 'futbol',
    # g
    'ga', 'gal', 'gallery', 'garden', 'gb', 'gbiz', 'gd', 'gdn',
    'ge', 'gent', 'gf', 'gg', 'ggee', 'gh', 'gi', 'gift', 'gifts',
    'gives', 'gl', 'glass', 'gle', 'global', 'globo', 'gm', 'gmail',
    'gmo', 'gmx', 'gn', 'goldpoint', 'goog', 'google', 'gop', 'gov',
    'gp', 'gq', 'gr', 'graphics', 'gratis', 'green', 'gripe', 'gs',
    'gt', 'gu', 'guide', 'guitars', 'guru', 'gw', 'gy',
    # h
    'hamburg', 'hangout', 'haus', 'healthcare', 'help', 'here',
    'hermes', 'hiphop', 'hiv', 'hk', 'hm', 'hn', 'holdings',
    'holiday', 'homes', 'horse', 'host', 'hosting', 'house', 'how',
    'hr', 'ht', 'hu',
    # i
    'ibm', 'id', 'ie', 'ifm', 'il', 'im', 'immo', 'immobilien', 'in',
    'industries', 'info', 'ing', 'ink', 'institute', 'insure', 'int',
    'international', 'investments', 'io', 'iq', 'ir', 'irish', 'is',
    'it', 'iwc',
    # j
    'jcb', 'je', 'jetzt', 'jm', 'jo', 'jobs', 'joburg', 'jp',
    'juegos',
    # k
    'kaufen', 'kddi', 'ke', 'kg', 'kh', 'ki', 'kim', 'kitchen',
    'kiwi', 'km', 'kn', 'koeln', 'kp', 'kr', 'krd', 'kred', 'kw',
    'ky', 'kyoto', 'kz',
    # l
    'la', 'lacaixa', 'land', 'lat', 'latrobe', 'lawyer', 'lb', 'lc',
    'lds', 'lease', 'legal', 'lgbt', 'li', 'lidl', 'life',
    'lighting', 'limited', 'limo', 'link', 'lk', 'loans',
    'localhost', 'london', 'lotte', 'lotto', 'lr', 'ls', 'lt',
    'ltda', 'lu', 'luxe', 'luxury', 'lv', 'ly',
    # m
    'ma', 'madrid', 'maison', 'management', 'mango', 'market',
    'marketing', 'marriott', 'mc', 'md', 'me', 'media', 'meet',
    'melbourne', 'meme', 'memorial', 'menu', 'mg', 'mh', 'miami',
    'mil', 'mini', 'mk', 'ml', 'mm', 'mn', 'mo', 'mobi', 'moda',
    'moe', 'monash', 'money', 'mormon', 'mortgage', 'moscow',
    'motorcycles', 'mov', 'mp', 'mq', 'mr', 'ms', 'mt', 'mu',
    'museum', 'mv', 'mw', 'mx', 'my', 'mz',
    # n
    'na', 'nagoya', 'name', 'navy', 'nc', 'ne', 'net', 'network',
    'neustar', 'new', 'nexus', 'nf', 'ng', 'ngo', 'nhk', 'ni',
    'nico', 'ninja', 'nl', 'no', 'np', 'nr', 'nra', 'nrw', 'ntt',
    'nu', 'nyc', 'nz',
    # o
    'okinawa', 'om', 'one', 'ong', 'onl', 'ooo', 'org', 'organic',
    'osaka', 'otsuka', 'ovh',
    # p
    'pa', 'paris', 'partners', 'parts', 'party', 'pe', 'pf', 'pg',
    'ph', 'pharmacy', 'photo', 'photography', 'photos', 'physio',
    'pics', 'pictures', 'pink', 'pizza', 'pk', 'pl', 'place',
    'plumbing', 'pm', 'pn', 'pohl', 'poker', 'porn', 'post', 'pr',
    'praxi', 'press', 'pro', 'prod', 'productions', 'prof',
    'properties', 'property', 'ps', 'pt', 'pub', 'pw', 'py',
    # q
    'qa', 'qpon', 'quebec',
    # r
    're', 'realtor', 'recipes', 'red', 'rehab', 'reise', 'reisen',
    'reit', 'ren', 'rentals', 'repair', 'report', 'republican',
    'rest', 'restaurant', 'reviews', 'rich', 'rio', 'rip', 'ro',
    'rocks', 'rodeo', 'rs', 'rsvp', 'ru', 'ruhr', 'rw', 'ryukyu',
    # s
    'sa', 'saarland', 'sale', 'samsung', 'sarl', 'saxo', 'sb', 'sc',
    'sca', 'scb', 'schmidt', 'school', 'schule', 'schwarz',
    'science', 'scot', 'sd', 'se', 'services', 'sew', 'sexy', 'sg',
    'sh', 'shiksha', 'shoes', 'shriram', 'si', 'singles', 'sj', 'sk',
    'sky', 'sl', 'sm', 'sn', 'so', 'social', 'software', 'sohu',
    'solar', 'solutions', 'soy', 'space', 'spiegel', 'sr', 'st',
    'style', 'su', 'supplies', 'supply', 'support', 'surf',
    'surgery', 'suzuki', 'sv', 'sx', 'sy', 'sydney', 'systems', 'sz',
    # t
    'taipei', 'tatar', 'tattoo', 'tax', 'tc', 'td', 'technology',
    'tel', 'temasek', 'tennis', 'tf', 'tg', 'th', 'tienda', 'tips',
    'tires', 'tirol', 'tj', 'tk', 'tl', 'tm', 'tn', 'to', 'today',
    'tokyo', 'tools', 'top', 'toshiba', 'town', 'toys', 'tp', 'tr',
    'trade', 'training', 'travel', 'trust', 'tt', 'tui', 'tv', 'tw',
    'tz',
    # u
    'ua', 'ug', 'uk', 'university', 'uno', 'uol', 'us', 'uy', 'uz',
    # v
    'va', 'vacations', 'vc', 've', 'vegas', 'ventures',
    'versicherung', 'vet', 'vg', 'vi', 'viajes', 'video', 'villas',
    'vision', 'vlaanderen', 'vn', 'vodka', 'vote', 'voting', 'voto',
    'voyage', 'vu',
    # w
    'wales', 'wang', 'watch', 'webcam', 'website', 'wed', 'wedding',
    'wf', 'whoswho', 'wien', 'wiki', 'williamhill', 'wme', 'work',
    'works', 'world', 'ws', 'wtc', 'wtf',
    # x
    'xn--1qqw23a', 'xn--3bst00m', 'xn--3ds443g', 'xn--3e0b707e',
    'xn--45brj9c', 'xn--45q11c', 'xn--4gbrim', 'xn--55qw42g',
    'xn--55qx5d', 'xn--6frz82g', 'xn--6qq986b3xl', 'xn--80adxhks',
    'xn--80ao21a', 'xn--80asehdb', 'xn--80aswg', 'xn--90a3ac',
    'xn--90ais', 'xn--b4w605ferd', 'xn--c1avg', 'xn--cg4bki',
    'xn--clchc0ea0b2g2a9gcd', 'xn--czr694b', 'xn--czrs0t',
    'xn--czru2d', 'xn--d1acj3b', 'xn--d1alf', 'xn--fiq228c5hs',
    'xn--fiq64b', 'xn--fiqs8s', 'xn--fiqz9s', 'xn--flw351e',
    'xn--fpcrj9c3d', 'xn--fzc2c9e2c', 'xn--gecrj9c', 'xn--h2brj9c',
    'xn--hxt814e', 'xn--i1b6b1a6a2e', 'xn--io0a7i', 'xn--j1amh',
    'xn--j6w193g', 'xn--kprw13d', 'xn--kpry57d', 'xn--kput3i',
    'xn--l1acc', 'xn--lgbbat1ad8j', 'xn--mgb9awbf',
    'xn--mgba3a4f16a', 'xn--mgbaam7a8h', 'xn--mgbab2bd',
    'xn--mgbayh7gpa', 'xn--mgbbh1a71e', 'xn--mgbc0a9azcg',
    'xn--mgberp4a5d4ar', 'xn--mgbx4cd0ab', 'xn--ngbc5azd',
    'xn--node', 'xn--nqv7f', 'xn--nqv7fs00ema', 'xn--o3cw4h',
    'xn--ogbpf8fl', 'xn--p1acf', 'xn--p1ai', 'xn--pgbs0dh',
    'xn--q9jyb4c', 'xn--qcka1pmc', 'xn--rhqv96g', 'xn--s9brj9c',
    'xn--ses554g', 'xn--unup4y', 'xn--vermgensberater-ctb',
    'xn--vermgensberatung-pwb', 'xn--vhquv', 'xn--wgbh1c',
    'xn--wgbl6a', 'xn--xhq521b', 'xn--xkc2al3hye2a',
    'xn--xkc2dl3a5ee0h', 'xn--yfro4i67o', 'xn--ygbi2ammx',
    'xn--zfr164b', 'xxx', 'xyz',
    # y
    'yachts', 'yandex', 'ye', 'yodobashi', 'yoga', 'yokohama',
    'youtube', 'yt',
    # z
    'za', 'zip', 'zm', 'zone', 'zuerich', 'zw'
]



# -*- coding: utf-8 -*-
"""
    emmett.cli
    ---------

    Provide command line tools for Emmett applications.

    :copyright: 2014 Giovanni Barillari

    Based on the code of Flask (http://flask.pocoo.org)
    :copyright: (c) 2014 by Armin Ronacher.

    :license: BSD-3-Clause
"""

import code
import os
import re
import ssl
import sys
import types

import click

from .__version__ import __version__ as fw_version
from ._internal import locate_app, get_app_module
from .asgi.loops import loops
from .asgi.protocols import protocols_http, protocols_ws
from .asgi.server import run as asgi_run
from .logger import LOG_LEVELS


def find_app_module():
    rv, files, dirs = None, [], []
    for path in os.listdir():
        if any(path.startswith(val) for val in [".", "test"]):
            continue
        if os.path.isdir(path):
            if not path.startswith("_"):
                dirs.append(path)
            continue
        _, ext = os.path.splitext(path)
        if ext == ".py":
            files.append(path)
    if "app.py" in files:
        rv = "app.py"
    elif "app" in dirs:
        rv = "app"
    elif "__init__.py" in files:
        rv = "__init__.py"
    elif len(files) == 1:
        rv = files[0]
    elif len(dirs) == 1:
        rv = dirs[1]
    else:
        modules = []
        for path in dirs:
            if os.path.exists(os.path.join(path, "__init__.py")):
                modules.append(path)
        if len(modules) == 1:
            rv = modules[0]
    return rv


def find_db(module, var_name=None):
    #: Given a module instance this tries to find the database instances
    #  in the module.
    if var_name:
        return [getattr(module, var_name)]

    from .orm import Database

    matches = [
        v for k, v in module.__dict__.items() if isinstance(v, Database)
    ]
    return matches


def get_import_components(path):
    return (re.split(r":(?![\\/])", path, 1) + [None])[:2]


def prepare_import(path):
    #: Given a path this will try to calculate the python path, add it
    #  to the search path and return the actual module name that is expected.
    path = os.path.realpath(path)

    fname, ext = os.path.splitext(path)
    if ext == ".py":
        path = fname
    if os.path.basename(path) == "__init__":
        path = os.path.dirname(path)

    module_name = []

    #: move up untile outside package
    while True:
        path, name = os.path.split(path)
        module_name.append(name)

        if not os.path.exists(os.path.join(path, "__init__.py")):
            break

    if sys.path[0] != path:
        sys.path.insert(0, path)

    return ".".join(module_name[::-1])


class ScriptInfo(object):
    def __init__(self, app_import_path=None, debug=None):
        #: The application import path
        self.app_import_path = app_import_path or os.environ.get("EMMETT_APP")
        #: The debug flag. If this is not None, the application will
        #  automatically have it's debug flag overridden with this value.
        self.debug = debug
        #: A dictionary with arbitrary data that can be associated with
        #  this script info.
        self.data = {}
        self._loaded_app = None
        self._loaded_ctx = None
        self.db_var_name = None

    def _get_import_name(self):
        if self.app_import_path:
            path, name = get_import_components(self.app_import_path)
        else:
            path, name = (find_app_module(), None)
        return prepare_import(path) if path else None, name

    def load_app(self):
        if self._loaded_app is not None:
            return self._loaded_app

        import_name, app_name = self._get_import_name()
        app = locate_app(import_name, app_name) if import_name else None

        if app is None:
            raise RuntimeError("Could not locate an Emmett application.")

        if self.debug is not None:
            app.debug = self.debug

        self._loaded_app = app
        return app

    def load_appctx(self):
        ctx = {}
        import_name, _ = self._get_import_name()
        mod = get_app_module(import_name)

        for key in set(mod.__dict__.keys()) - {"__builtins__"}:
            value = mod.__dict__[key]
            if isinstance(value, types.FunctionType):
                continue
            ctx[key] = value

        self._loaded_ctx = ctx
        return ctx

    def load_db(self):
        import_name, _ = self._get_import_name()
        mod = get_app_module(import_name)
        return find_db(mod, self.db_var_name)


pass_script_info = click.make_pass_decorator(ScriptInfo)


def set_app_value(ctx, param, value):
    ctx.ensure_object(ScriptInfo).app_import_path = value


app_option = click.Option(
    ['-a', '--app'],
    help='The application to run',
    callback=set_app_value,
    is_eager=True
)


class EmmettGroup(click.Group):
    def __init__(
        self,
        add_default_commands=True,
        add_app_option=True,
        add_debug_option=True,
        **extra
    ):
        params = list(extra.pop('params', None) or ())
        if add_app_option:
            params.append(app_option)
        #if add_debug_option:
        #    params.append(debug_option)

        click.Group.__init__(self, params=params, **extra)
        #self.create_app = create_app

        if add_default_commands:
            self.add_command(develop_command)
            self.add_command(shell_command)
            self.add_command(routes_command)
            self.add_command(serve_command)

    def list_commands(self, ctx):
        rv = super(EmmettGroup, self).list_commands(ctx)
        info = ctx.ensure_object(ScriptInfo)
        try:
            rv = rv + info.load_app().cli.list_commands(ctx)
        except Exception:
            pass
        return rv

    def get_command(self, ctx, name):
        # We load built-in commands first as these should always be the
        # same no matter what the app does.  If the app does want to
        # override this it needs to make a custom instance of this group
        # and not attach the default commands.
        #
        # This also means that the script stays functional in case the
        # application completely fails.
        rv = click.Group.get_command(self, ctx, name)
        if rv is not None:
            return rv

        info = ctx.ensure_object(ScriptInfo)
        try:
            rv = info.load_app().cli.get_command(ctx, name)
            if rv is not None:
                return rv
        except Exception:
            pass

    def main(self, *args, **kwargs):
        obj = kwargs.get('obj')
        if obj is None:
            obj = ScriptInfo()
        kwargs['obj'] = obj
        return super().main(*args, **kwargs)


@click.command('develop', short_help='Runs a development server.')
@click.option(
    '--host', '-h', default='127.0.0.1', help='The interface to bind to.')
@click.option(
    '--port', '-p', type=int, default=8000, help='The port to bind to.')
@click.option(
    '--loop', type=click.Choice(loops.keys()), default='auto',
    help='Event loop implementation.')
@click.option(
    '--http-protocol', type=click.Choice(protocols_http.keys()),
    default='auto', help='HTTP protocol implementation.')
@click.option(
    '--ws-protocol', type=click.Choice(protocols_ws.keys()),
    default='auto', help='Websocket protocol implementation.')
@click.option(
    '--ssl-certfile', type=str, default=None, help='SSL certificate file')
@click.option(
    '--ssl-keyfile', type=str, default=None, help='SSL key file')
@click.option(
    '--ssl-cert-reqs', type=int, default=ssl.CERT_NONE,
    help='Whether client certificate is required (see ssl module)')
@click.option(
    '--ssl-ca-certs', type=str, default=None, help='CA certificates file')
@click.option(
    '--reloader/--no-reloader', is_flag=True, default=True,
    help='Runs with reloader.')
@pass_script_info
def develop_command(
    info, host, port,
    loop, http_protocol, ws_protocol,
    ssl_certfile, ssl_keyfile, ssl_cert_reqs, ssl_ca_certs,
    reloader
):
    os.environ["EMMETT_RUN_ENV"] = 'true'
    app_target = info._get_import_name()

    if os.environ.get('EMMETT_RUN_MAIN') != 'true':
        click.echo(
            ' '.join([
                "> Starting Emmett development server on app",
                click.style(app_target[0], fg="cyan", bold=True)
            ])
        )
        click.echo(
            ' '.join([
                click.style("> Emmett application", fg="green"),
                click.style(app_target[0], fg="cyan", bold=True),
                click.style("running on", fg="green"),
                click.style(f"http://{host}:{port}", fg="cyan"),
                click.style("(press CTRL+C to quit)", fg="green")
            ])
        )

    if reloader:
        from ._reloader import run_with_reloader
        runner = run_with_reloader
    else:
        runner = asgi_run

    runner(
        app_target,
        host,
        port,
        loop=loop,
        proto_http=http_protocol,
        proto_ws=ws_protocol,
        log_level='debug',
        access_log=True,
        ssl_certfile=ssl_certfile,
        ssl_keyfile=ssl_keyfile,
        ssl_cert_reqs=ssl_cert_reqs,
        ssl_ca_certs=ssl_ca_certs
    )


@click.command('serve', short_help='Serve the app.')
@click.option(
    '--host', '-h', default='0.0.0.0', help='The interface to bind to.')
@click.option(
    '--port', '-p', type=int, default=8000, help='The port to bind to.')
@click.option(
    "--workers", type=int, default=1, help="Number of worker processes. Defaults to 1.")
@click.option(
    '--loop', type=click.Choice(loops.keys()), default='auto',
    help='Event loop implementation.')
@click.option(
    '--http-protocol', type=click.Choice(protocols_http.keys()),
    default='auto', help='HTTP protocol implementation.')
@click.option(
    '--ws-protocol', type=click.Choice(protocols_ws.keys()),
    default='auto', help='Websocket protocol implementation.')
@click.option(
    '--log-level', type=click.Choice(LOG_LEVELS.keys()), default='info',
    help='Logging level.')
@click.option(
    '--access-log/--no-access-log', is_flag=True, default=True,
    help='Enable/Disable access log.')
@click.option(
    '--proxy-headers/--no-proxy-headers', is_flag=True, default=False,
    help='Enable/Disable proxy headers.')
@click.option(
    '--proxy-trust-ips', type=str, default=None,
    help='Comma seperated list of IPs to trust with proxy headers')
@click.option(
    '--max-concurrency', type=int,
    help='The maximum number of concurrent connections.')
@click.option(
    '--backlog', type=int, default=2048,
    help='Maximum number of connections to hold in backlog')
@click.option(
    '--keep-alive-timeout', type=int, default=0,
    help='Keep alive timeout for connections.')
@click.option(
    '--ssl-certfile', type=str, default=None, help='SSL certificate file')
@click.option(
    '--ssl-keyfile', type=str, default=None, help='SSL key file')
@click.option(
    '--ssl-cert-reqs', type=int, default=ssl.CERT_NONE,
    help='Whether client certificate is required (see ssl module)')
@click.option(
    '--ssl-ca-certs', type=str, default=None, help='CA certificates file')
@pass_script_info
def serve_command(
    info, host, port, workers,
    loop, http_protocol, ws_protocol,
    log_level, access_log,
    proxy_headers, proxy_trust_ips,
    max_concurrency, backlog, keep_alive_timeout,
    ssl_certfile, ssl_keyfile, ssl_cert_reqs, ssl_ca_certs
):
    app_target = info._get_import_name()
    asgi_run(
        app_target,
        host=host, port=port,
        loop=loop, proto_http=http_protocol, proto_ws=ws_protocol,
        log_level=log_level, access_log=access_log,
        proxy_headers=proxy_headers, proxy_trust_ips=proxy_trust_ips,
        workers=workers,
        limit_concurrency=max_concurrency,
        backlog=backlog,
        timeout_keep_alive=keep_alive_timeout,
        ssl_certfile=ssl_certfile,
        ssl_keyfile=ssl_keyfile,
        ssl_cert_reqs=ssl_cert_reqs,
        ssl_ca_certs=ssl_ca_certs
    )


@click.command('shell', short_help='Runs a shell in the app context.')
@pass_script_info
def shell_command(info):
    os.environ['EMMETT_CLI_ENV'] = 'true'
    ctx = info.load_appctx()
    app = info.load_app()
    banner = 'Python %s on %s\nEmmett %s shell on app: %s' % (
        sys.version,
        sys.platform,
        fw_version,
        app.import_name
    )
    code.interact(banner=banner, local=app.make_shell_context(ctx))


@click.command('routes', short_help='Display the app routing table.')
@pass_script_info
def routes_command(info):
    app = info.load_app()
    click.echo(
        "".join([
            "> Routing table for Emmett application ",
            click.style(app.import_name, fg="cyan", bold=True),
            ":"
        ])
    )
    for route in app._router_http._routes_str.values():
        click.echo(route)
    for route in app._router_ws._routes_str.values():
        click.echo(route)


cli = EmmettGroup(help="")


def set_db_value(ctx, param, value):
    ctx.ensure_object(ScriptInfo).db_var_name = value


@cli.group('migrations', short_help='Runs migration operations.')
@click.option(
    '--db', help='The db instance to use', callback=set_db_value,
    is_eager=True)
def migrations_cli(db):
    pass


@migrations_cli.command(
    'status', short_help='Shows current database revision.')
@click.option('--verbose', '-v', default=False, is_flag=True)
@pass_script_info
def migrations_status(info, verbose):
    from .orm.migrations.commands import status
    app = info.load_app()
    dbs = info.load_db()
    status(app, dbs, verbose)


@migrations_cli.command('history', short_help="Shows migrations history.")
@click.option('--range', '-r', default=None)
@click.option('--verbose', '-v', default=False, is_flag=True)
@pass_script_info
def migrations_history(info, range, verbose):
    from .orm.migrations.commands import history
    app = info.load_app()
    dbs = info.load_db()
    history(app, dbs, range, verbose)


@migrations_cli.command(
    'generate', short_help='Generate a new migration from application models.')
@click.option('--message', '-m', default='Generated migration',
              help='The description for the new migration.')
@click.option('-head', default='head', help='The migration to generate from')
@pass_script_info
def migrations_generate(info, message, head):
    from .orm.migrations.commands import generate
    app = info.load_app()
    dbs = info.load_db()
    generate(app, dbs, message, head)


@migrations_cli.command('new', short_help='Generate a new empty migration.')
@click.option('--message', '-m', default='New migration',
              help='The description for the new migration.')
@click.option('-head', default='head', help='The migration to generate from')
@pass_script_info
def migrations_new(info, message, head):
    from .orm.migrations.commands import new
    app = info.load_app()
    dbs = info.load_db()
    new(app, dbs, message, head)


@migrations_cli.command(
    'up', short_help='Upgrades the database to the selected migration.'
)
@click.option('--revision', '-r', default='head', help='The migration to upgrade to.')
@click.option(
    '--dry-run',
    default=False,
    is_flag=True,
    help='Only print SQL instructions, without actually applying the migration.'
)
@pass_script_info
def migrations_up(info, revision, dry_run):
    from .orm.migrations.commands import up
    app = info.load_app()
    dbs = info.load_db()
    up(app, dbs, revision, dry_run)


@migrations_cli.command(
    'down', short_help='Downgrades the database to the selected migration.'
)
@click.option('--revision', '-r', required=True, help='The migration to downgrade to.')
@click.option(
    '--dry-run',
    default=False,
    is_flag=True,
    help='Only print SQL instructions, without actually applying the migration.'
)
@pass_script_info
def migrations_down(info, revision, dry_run):
    from .orm.migrations.commands import down
    app = info.load_app()
    dbs = info.load_db()
    down(app, dbs, revision, dry_run)


def main(as_module=False):
    cli.main(prog_name="python -m emmett" if as_module else None)


if __name__ == '__main__':
    main(as_module=True)



# -*- coding: utf-8 -*-
"""
    emmett.orm.migrations.operations
    --------------------------------

    Provides operations handlers for migrations.

    :copyright: 2014 Giovanni Barillari

    Based on the code of Alembic (https://bitbucket.org/zzzeek/alembic)
    :copyright: (c) 2009-2015 by Michael Bayer

    :license: BSD-3-Clause
"""

import re
from .base import Migration, Column
from .helpers import DEFAULT_VALUE


class Operation(object):
    def _env_load_(self, engine):
        self.engine = engine

    def run(self):
        pass


class OpContainer(Operation):
    #: represent a sequence of operations
    def __init__(self, ops=()):
        self.ops = ops

    def is_empty(self):
        return not self.ops

    def as_diffs(self):
        return list(OpContainer._ops_as_diffs(self))

    @classmethod
    def _ops_as_diffs(cls, migrations):
        for op in migrations.ops:
            if hasattr(op, 'ops'):
                for sub_op in cls._ops_as_diffs(op):
                    yield sub_op
            else:
                yield op.to_diff_tuple()


class ModifyTableOps(OpContainer):
    #: a sequence of operations that all apply to a single Table
    def __init__(self, table_name, ops):
        super(ModifyTableOps, self).__init__(ops)
        self.table_name = table_name

    def reverse(self):
        return ModifyTableOps(
            self.table_name,
            ops=list(reversed(
                [op.reverse() for op in self.ops]
            ))
        )


class UpgradeOps(OpContainer):
    #: contains a sequence of operations that would apply during upgrade
    def __init__(self, ops=(), upgrade_token="upgrades"):
        super(UpgradeOps, self).__init__(ops=ops)
        self.upgrade_token = upgrade_token

    def reverse(self):
        return DowngradeOps(
            ops=list(reversed(
                [op.reverse() for op in self.ops]
            ))
        )


class DowngradeOps(OpContainer):
    #: contains a sequence of operations that would apply during downgrade
    def __init__(self, ops=(), downgrade_token="downgrades"):
        super(DowngradeOps, self).__init__(ops=ops)
        self.downgrade_token = downgrade_token

    def reverse(self):
        return UpgradeOps(
            ops=list(reversed(
                [op.reverse() for op in self.ops]
            ))
        )


class MigrationOp(Operation):
    def __init__(self, rev_id, upgrade_ops, downgrade_ops, message=None,
                 head=None, splice=None):
        self.rev_id = rev_id
        self.message = message
        self.head = head
        self.splice = splice
        self.upgrade_ops = upgrade_ops
        self.downgrade_ops = downgrade_ops


@Migration.register_operation("create_table")
class CreateTableOp(Operation):
    def __init__(self, table_name, columns, _orig_table=None, **kw):
        self.table_name = table_name
        self.columns = columns
        self.kw = kw
        self._orig_table = _orig_table

    def reverse(self):
        return DropTableOp.from_table(self.to_table())

    def to_diff_tuple(self):
        return ("add_table", self.to_table())

    @classmethod
    def from_table(cls, table):
        return cls(
            table.name,
            [table[colname] for colname in table.fields],
            _orig_table=table
        )

    def to_table(self, migration_context=None):
        if self._orig_table is not None:
            return self._orig_table
        from .generation import MetaTable
        return MetaTable(
            self.table_name, self.columns, **self.kw
        )

    @classmethod
    def create_table(cls, table_name, *columns, **kw):
        return cls(table_name, columns, **kw)

    def run(self):
        self.engine.create_table(self.table_name, self.columns, **self.kw)


@Migration.register_operation("drop_table")
class DropTableOp(Operation):
    def __init__(self, table_name, table_kw=None, _orig_table=None):
        self.table_name = table_name
        self.table_kw = table_kw or {}
        self._orig_table = _orig_table

    def to_diff_tuple(self):
        return ("remove_table", self.to_table())

    def reverse(self):
        if self._orig_table is None:
            raise ValueError(
                "operation is not reversible; "
                "original table is not present")
        return CreateTableOp.from_table(self._orig_table)

    @classmethod
    def from_table(cls, table):
        return cls(table.name, _orig_table=table)

    def to_table(self):
        if self._orig_table is not None:
            return self._orig_table
        from .generation import MetaTable
        return MetaTable(
            self.table_name,
            **self.table_kw)

    @classmethod
    def drop_table(cls, table_name, **kw):
        return cls(table_name, table_kw=kw)

    def run(self):
        self.engine.drop_table(self.table_name)


class AlterTableOp(Operation):
    def __init__(self, table_name):
        self.table_name = table_name


@Migration.register_operation("rename_table")
class RenameTableOp(AlterTableOp):
    def __init__(self, old_table_name, new_table_name):
        super(RenameTableOp, self).__init__(old_table_name)
        self.new_table_name = new_table_name

    @classmethod
    def rename_table(cls, old_table_name, new_table_name):
        return cls(old_table_name, new_table_name)

    def run(self):
        raise NotImplementedError(
            'Table renaming is currently not supported.'
        )


@Migration.register_operation("add_column")
class AddColumnOp(AlterTableOp):
    def __init__(self, table_name, column):
        super(AddColumnOp, self).__init__(table_name)
        self.column = column

    def reverse(self):
        return DropColumnOp.from_column_and_tablename(
            self.table_name, self.column)

    def to_diff_tuple(self):
        return ("add_column", self.table_name, self.column)

    def to_column(self):
        return self.column

    @classmethod
    def from_column_and_tablename(cls, tname, col):
        return cls(tname, col)

    @classmethod
    def add_column(cls, table_name, column):
        return cls(table_name, column)

    def run(self):
        self.engine.add_column(self.table_name, self.column)


@Migration.register_operation("drop_column")
class DropColumnOp(AlterTableOp):
    def __init__(self, table_name, column_name, _orig_column=None, **kw):
        super(DropColumnOp, self).__init__(table_name)
        self.column_name = column_name
        self.kw = kw
        self._orig_column = _orig_column

    def to_diff_tuple(self):
        return ("remove_column", self.table_name, self.to_column())

    def reverse(self):
        if self._orig_column is None:
            raise ValueError(
                "operation is not reversible; "
                "original column is not present")

        return AddColumnOp.from_column_and_tablename(
            self.table_name, self._orig_column)

    @classmethod
    def from_column_and_tablename(cls, tname, col):
        return cls(tname, col.name, _orig_column=col)

    def to_column(self):
        if self._orig_column is not None:
            return self._orig_column
        return Column(self.column_name, **self.kw)

    @classmethod
    def drop_column(cls, table_name, column_name, **kw):
        return cls(table_name, column_name, **kw)

    def run(self):
        self.engine.drop_column(self.table_name, self.column_name)


@Migration.register_operation("alter_column")
class AlterColumnOp(AlterTableOp):
    def __init__(
            self, table_name, column_name,
            existing_type=None,
            existing_length=None,
            existing_default=None,
            existing_notnull=None,
            modify_notnull=None,
            modify_default=DEFAULT_VALUE,
            modify_name=None,
            modify_type=None,
            modify_length=None,
            **kw

    ):
        super(AlterColumnOp, self).__init__(table_name)
        self.column_name = column_name
        self.existing_type = existing_type
        self.existing_length = existing_length
        self.existing_default = existing_default
        self.existing_notnull = existing_notnull
        self.modify_notnull = modify_notnull
        self.modify_default = modify_default
        self.modify_name = modify_name
        self.modify_type = modify_type
        self.modify_length = modify_length
        self.kw = kw

    def to_diff_tuple(self):
        col_diff = []
        tname, cname = self.table_name, self.column_name

        if self.modify_type is not None:
            col_diff.append(
                (
                    "modify_type", tname, cname, {
                        "existing_length": self.existing_length,
                        "existing_notnull": self.existing_notnull,
                        "existing_default": self.existing_default},
                    self.existing_type,
                    self.modify_type
                )
            )

        if self.modify_length is not None:
            col_diff.append(
                (
                    "modify_length", tname, cname, {
                        "existing_type": self.existing_type,
                        "existing_notnull": self.existing_notnull,
                        "existing_default": self.existing_default},
                    self.existing_length,
                    self.modify_length
                )
            )

        if self.modify_notnull is not None:
            col_diff.append(
                (
                    "modify_notnull", tname, cname, {
                        "existing_type": self.existing_type,
                        "existing_default": self.existing_default},
                    self.existing_notnull,
                    self.modify_notnull
                )
            )

        if self.modify_default is not DEFAULT_VALUE:
            col_diff.append(
                (
                    "modify_default", tname, cname, {
                        "existing_notnull": self.existing_notnull,
                        "existing_type": self.existing_type},
                    self.existing_default,
                    self.modify_default
                )
            )

        return col_diff

    def has_changes(self):
        hc = (
            self.modify_notnull is not None or
            self.modify_default is not DEFAULT_VALUE or
            self.modify_type is not None or
            self.modify_length is not None
        )
        if hc:
            return True
        for kw in self.kw:
            if kw.startswith('modify_'):
                return True
        return False

    def reverse(self):
        kw = self.kw.copy()
        kw['existing_type'] = self.existing_type
        kw['existing_length'] = self.existing_length
        kw['existing_notnull'] = self.existing_notnull
        kw['existing_default'] = self.existing_default
        if self.modify_type is not None:
            kw['modify_type'] = self.modify_type
        if self.modify_length is not None:
            kw['modify_length'] = self.modify_length
        if self.modify_notnull is not None:
            kw['modify_notnull'] = self.modify_notnull
        if self.modify_default is not DEFAULT_VALUE:
            kw['modify_default'] = self.modify_default

        all_keys = set(m.group(1) for m in [
            re.match(r'^(?:existing_|modify_)(.+)$', k)
            for k in kw
        ] if m)

        for k in all_keys:
            if 'modify_%s' % k in kw:
                swap = kw['existing_%s' % k]
                kw['existing_%s' % k] = kw['modify_%s' % k]
                kw['modify_%s' % k] = swap

        return self.__class__(
            self.table_name, self.column_name, **kw
        )

    @classmethod
    def alter_column(
        cls, table_name, column_name,
        notnull=None,
        default=DEFAULT_VALUE,
        new_column_name=None,
        type=None,
        length=None,
        existing_type=None,
        existing_length=None,
        existing_default=None,
        existing_notnull=None,
        **kw
    ):
        return cls(
            table_name, column_name,
            existing_type=existing_type,
            existing_length=existing_length,
            existing_default=existing_default,
            existing_notnull=existing_notnull,
            modify_name=new_column_name,
            modify_type=type,
            modify_length=length,
            modify_default=default,
            modify_notnull=notnull,
            **kw
        )

    def run(self):
        self.engine.alter_column(
            self.table_name, self.column_name, self.to_diff_tuple())


@Migration.register_operation("create_index")
class CreateIndexOp(Operation):
    def __init__(
        self, index_name, table_name, fields=[], expressions=[], unique=False,
        _orig_index=None, **kw
    ):
        self.index_name = index_name
        self.table_name = table_name
        self.fields = fields
        self.expressions = expressions
        self.unique = unique
        self.kw = kw
        self._orig_index = _orig_index

    def reverse(self):
        return DropIndexOp.from_index(self.to_index())

    def to_diff_tuple(self):
        return ("create_index", self.to_index())

    @classmethod
    def from_index(cls, index):
        return cls(
            index.name, index.table_name, index.fields, index.expressions,
            index.unique, _orig_index=index, **index.kw
        )

    def to_index(self):
        if self._orig_index is not None:
            return self._orig_index
        from .generation import MetaIndex
        return MetaIndex(
            self.table_name, self.index_name, self.fields, self.expressions,
            self.unique, **self.kw)

    @classmethod
    def create_index(
        cls, index_name, table_name, fields=[], expressions=[], unique=False,
        **kw
    ):
        return cls(index_name, table_name, fields, expressions, unique, **kw)

    def run(self):
        self.engine.create_index(
            self.index_name, self.table_name, self.fields, self.expressions,
            self.unique, **self.kw)


@Migration.register_operation("drop_index")
class DropIndexOp(Operation):
    def __init__(self, index_name, table_name=None, _orig_index=None):
        self.index_name = index_name
        self.table_name = table_name
        self._orig_index = _orig_index

    def to_diff_tuple(self):
        return ("remove_index", self.to_index())

    def reverse(self):
        if self._orig_index is None:
            raise ValueError(
                "operation is not reversible; "
                "original index is not present")
        return CreateIndexOp.from_index(self._orig_index)

    @classmethod
    def from_index(cls, index):
        return cls(index.name, index.table_name, index)

    def to_index(self):
        if self._orig_index is not None:
            return self._orig_index
        from .generation import MetaIndex
        return MetaIndex(self.table_name, self.index_name, [], [], False)

    @classmethod
    def drop_index(cls, index_name, table_name):
        return cls(index_name, table_name)

    def run(self):
        self.engine.drop_index(self.index_name, self.table_name)


# @Migration.register_operation("execute")
# class ExecuteSQLOp(Operation):
#     def __init__(self, sqltext):
#         self.sqltext = sqltext

#     def run(self):
#         pass



# -*- coding: utf-8 -*-
"""
    emmett.asgi.loops.auto
    ----------------------

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from . import loops


@loops.register('auto')
def build_auto_loop():
    if 'uvloop' in loops:
        return loops.get('uvloop')
    return loops.get('asyncio')



# -*- coding: utf-8 -*-
"""
    emmett.asgi.protocols.http.h11
    ------------------------------

    Provides HTTP h11 protocol implementation

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import asyncio

import h11

from uvicorn.protocols.http.h11_impl import (
    HIGH_WATER_LIMIT,
    STATUS_PHRASES,
    H11Protocol as _H11Protocol,
    RequestResponseCycle,
    service_unavailable,
    unquote
)

from . import protocols
from .h2 import H2Protocol


@protocols.register("h11")
class H11Protocol(_H11Protocol):
    alpn_protocols = ["h2", "http/1.1"]
    h2_protocol_class = H2Protocol

    def handle_upgrade(self, event: h11.Request):
        upgrade_value = None
        for name, value in self.headers:
            if name == b"upgrade":
                upgrade_value = value.lower()
                break

        if upgrade_value == b"websocket" and self.ws_protocol_class:
            self.connections.discard(self)
            output = [event.method, b" ", event.target, b" HTTP/1.1\r\n"]
            for name, value in self.headers:
                output += [name, b": ", value, b"\r\n"]
            output.append(b"\r\n")
            protocol = self.ws_protocol_class(
                config=self.config,
                server_state=self.server_state,
                on_connection_lost=self.on_connection_lost
            )
            protocol.connection_made(self.transport)
            protocol.data_received(b"".join(output))
            self.transport.set_protocol(protocol)
        elif upgrade_value == b"h2c":
            self.connections.discard(self)
            self.transport.write(
                self.conn.send(
                    h11.InformationalResponse(
                        status_code=101,
                        headers=self.headers
                    )
                )
            )
            protocol = self.h2_protocol_class(
                config=self.config,
                server_state=self.server_state,
                on_connection_lost=self.on_connection_lost,
                _loop=self.loop
            )
            protocol.handle_upgrade_from_h11(self.transport, event, self.headers)
            self.transport.set_protocol(protocol)
        else:
            msg = "Unsupported upgrade request."
            self.logger.warning(msg)
            reason = STATUS_PHRASES[400]
            headers = [
                (b"content-type", b"text/plain; charset=utf-8"),
                (b"connection", b"close"),
            ]
            event = h11.Response(status_code=400, headers=headers, reason=reason)
            output = self.conn.send(event)
            self.transport.write(output)
            event = h11.Data(data=b"Unsupported upgrade request.")
            output = self.conn.send(event)
            self.transport.write(output)
            event = h11.EndOfMessage()
            output = self.conn.send(event)
            self.transport.write(output)
            self.transport.close()

    def handle_h2_assumed(self):
        self.connections.discard(self)
        protocol = self.h2_protocol_class(
            config=self.config,
            server_state=self.server_state,
            _loop=self.loop
        )
        protocol.connection_made(self.transport)
        protocol.data_received(
            b"PRI * HTTP/2.0\r\n\r\n" + self.conn.trailing_data[0]
        )
        self.transport.set_protocol(protocol)

    def handle_events(self):
        while True:
            try:
                event = self.conn.next_event()
            except h11.RemoteProtocolError as exc:
                msg = "Invalid HTTP request received."
                self.logger.warning(msg, exc_info=exc)
                self.transport.close()
                return
            event_type = type(event)

            if event_type is h11.NEED_DATA:
                break

            elif event_type is h11.PAUSED:
                # This case can occur in HTTP pipelining, so we need to
                # stop reading any more data, and ensure that at the end
                # of the active request/response cycle we handle any
                # events that have been buffered up.
                self.flow.pause_reading()
                break

            elif event_type is h11.Request:
                self.headers, upgrade_value = [], None
                for name, value in event.headers:
                    lname = name.lower()
                    self.headers.append((lname, value))
                    if lname == b"upgrade":
                        upgrade_value = value

                if upgrade_value:
                    self.handle_upgrade(event)
                    return
                elif (
                    event.method == b"PRI" and
                    event.target == b"*" and
                    event.http_version == b"2.0"
                ):
                    self.handle_h2_assumed()
                    return

                raw_path, _, query_string = event.target.partition(b"?")
                self.scope = {
                    "type": "http",
                    "asgi": {
                        "version": self.config.asgi_version,
                        "spec_version": "2.1",
                    },
                    "http_version": event.http_version.decode("ascii"),
                    "server": self.server,
                    "client": self.client,
                    "scheme": self.scheme,
                    "method": event.method.decode("ascii"),
                    "root_path": self.root_path,
                    "path": unquote(raw_path.decode("ascii")),
                    "raw_path": raw_path,
                    "query_string": query_string,
                    "headers": self.headers
                }

                # Handle 503 responses when 'limit_concurrency' is exceeded.
                if self.limit_concurrency is not None and (
                    len(self.connections) >= self.limit_concurrency
                    or len(self.tasks) >= self.limit_concurrency
                ):
                    app = service_unavailable
                    message = "Exceeded concurrency limit."
                    self.logger.warning(message)
                else:
                    app = self.app

                self.cycle = RequestResponseCycle(
                    scope=self.scope,
                    conn=self.conn,
                    transport=self.transport,
                    flow=self.flow,
                    logger=self.logger,
                    access_logger=self.access_logger,
                    access_log=self.access_log,
                    default_headers=self.default_headers,
                    message_event=asyncio.Event(),
                    on_response=self.on_response_complete,
                )
                task = self.loop.create_task(self.cycle.run_asgi(app))
                task.add_done_callback(self.tasks.discard)
                self.tasks.add(task)

            elif event_type is h11.Data:
                if self.conn.our_state is h11.DONE:
                    continue
                self.cycle.body += event.data
                if len(self.cycle.body) > HIGH_WATER_LIMIT:
                    self.flow.pause_reading()
                self.cycle.message_event.set()

            elif event_type is h11.EndOfMessage:
                if self.conn.our_state is h11.DONE:
                    self.transport.resume_reading()
                    self.conn.start_next_cycle()
                    continue
                self.cycle.more_body = False
                self.cycle.message_event.set()



# -*- coding: utf-8 -*-
"""
    emmett.orm.base
    ---------------

    Provides base pyDAL implementation for Emmett.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import copyreg
import os
import threading

from functools import wraps
from pydal import DAL as _pyDAL
from pydal._globals import THREAD_LOCAL

from ..datastructures import sdict
from ..extensions import Signals
from ..pipeline import Pipe
from ..security import uuid as _uuid
from ..serializers import _json_default, xml
from .adapters import patch_adapter
from .objects import Table, Field, Set, Row, Rows
from .helpers import ConnectionContext, TimingHandler
from .models import MetaModel, Model
from .transactions import _atomic, _transaction, _savepoint


class DatabasePipe(Pipe):
    def __init__(self, db):
        self.db = db

    async def open(self):
        await self.db.connection_open_loop()

    async def on_pipe_success(self):
        self.db.commit()

    async def on_pipe_failure(self):
        self.db.rollback()

    async def close(self):
        await self.db.connection_close_loop()


class Database(_pyDAL):
    serializers = {'json': _json_default, 'xml': xml}
    logger = None
    uuid = lambda x: _uuid()

    record_operators = {}
    execution_handlers = []

    Rows = Rows
    Row = Row

    _cls_global_lock_ = threading.RLock()

    @staticmethod
    def uri_from_config(config=None):
        if config is None or config.adapter is None:
            config = sdict(adapter="sqlite", host="dummy.db")
        if config.adapter == "<zombie>":
            return config.adapter
        if config.adapter == "sqlite" and config.host == "memory":
            return config.adapter + ":" + config.host
        uri = config.adapter + "://"
        if config.user:
            uri = uri + config.user + ":" + config.password + "@"
        uri = uri + config.host
        if config.database:
            uri += "/" + config.database
        return uri

    def __new__(cls, app, *args, **kwargs):
        config = kwargs.get('config', sdict()) or app.config.db
        uri = config.uri or Database.uri_from_config(config)
        return super(Database, cls).__new__(cls, uri, *args, **kwargs)

    def __init__(
        self,
        app,
        config=None,
        pool_size=None,
        keep_alive_timeout=3600,
        connect_timeout=60,
        folder=None,
        **kwargs
    ):
        app.send_signal(Signals.before_database)
        self.logger = app.log
        config = config or app.config.db
        if not config.uri:
            config.uri = self.uri_from_config(config)
        if not config.migrations_folder:
            config.migrations_folder = 'migrations'
        self.config = config
        self._auto_migrate = self.config.get(
            'auto_migrate', kwargs.pop('auto_migrate', False))
        self._auto_connect = self.config.get(
            'auto_connect', kwargs.pop('auto_connect', None))
        self._use_bigint_on_id_fields = self.config.get(
            'big_id_fields', kwargs.pop('big_id_fields', False))
        #: load config data
        kwargs['check_reserved'] = self.config.check_reserved or \
            kwargs.get('check_reserved', None)
        kwargs['migrate'] = self._auto_migrate
        kwargs['driver_args'] = self.config.driver_args or \
            kwargs.get('driver_args', None)
        kwargs['adapter_args'] = self.config.adapter_args or \
            kwargs.get('adapter_args', None)
        if self._auto_connect is not None:
            kwargs['do_connect'] = self._auto_connect
        else:
            kwargs['do_connect'] = os.environ.get('EMMETT_CLI_ENV') == 'true'
        if self._use_bigint_on_id_fields:
            kwargs['bigint_id'] = True
        #: set directory
        folder = folder or 'databases'
        folder = os.path.join(app.root_path, folder)
        if self._auto_migrate:
            with self._cls_global_lock_:
                if not os.path.exists(folder):
                    os.mkdir(folder)
        #: set pool_size
        pool_size = self.config.pool_size or pool_size or 5
        self._keep_alive_timeout = (
            keep_alive_timeout if self.config.keep_alive_timeout is None
            else self.config.keep_alive_timeout)
        self._connect_timeout = (
            connect_timeout if self.config.connect_timeout is None
            else self.config.connect_timeout)
        #: add timings storage if requested
        if config.store_execution_timings:
            self.execution_handlers.append(TimingHandler)
        #: finally setup pyDAL instance
        super(Database, self).__init__(
            self.config.uri, pool_size, folder, **kwargs)
        patch_adapter(self._adapter)
        Model._init_inheritable_dicts_()
        app.send_signal(Signals.after_database, database=self)

    @property
    def pipe(self):
        return DatabasePipe(self)

    @property
    def execution_timings(self):
        return getattr(THREAD_LOCAL, '_emtdal_timings_', [])

    def connection_open(self, with_transaction=True, reuse_if_open=True):
        return self._adapter.reconnect(
            with_transaction=with_transaction, reuse_if_open=reuse_if_open)

    def connection_close(self):
        self._adapter.close()

    def connection(
        self,
        with_transaction: bool = True,
        reuse_if_open: bool = True
    ) -> ConnectionContext:
        return ConnectionContext(
            self,
            with_transaction=with_transaction,
            reuse_if_open=reuse_if_open
        )

    def connection_open_loop(self, with_transaction=True, reuse_if_open=True):
        return self._adapter.reconnect_loop(
            with_transaction=with_transaction, reuse_if_open=reuse_if_open)

    def connection_close_loop(self):
        return self._adapter.close_loop()

    def define_models(self, *models):
        if len(models) == 1 and isinstance(models[0], (list, tuple)):
            models = models[0]
        if self._auto_migrate and not self._do_connect:
            self.connection_open()
        for model in models:
            if not hasattr(self, model.__name__):
                # store db instance inside model
                model.db = self
                # init model
                obj = model()
                obj._define_props_()
                obj._define_relations_()
                obj._define_virtuals_()
                obj._build_rowclass_()
                # define table and store in model
                args = dict(
                    migrate=obj.migrate,
                    format=obj.format,
                    table_class=Table
                )
                model.table = self.define_table(
                    obj.tablename, *obj.fields, **args
                )
                model.table._model_ = obj
                # load user's definitions
                obj._define_()
                # set reference in db for model name
                self.__setattr__(model.__name__, obj.table)
        if self._auto_migrate and not self._do_connect:
            self.connection_close()

    def where(self, query=None, ignore_common_filters=None, model=None):
        q = None
        if isinstance(query, Table):
            q = self._adapter.id_query(query)
        elif isinstance(query, Field):
            q = (query != None)
        elif isinstance(query, dict):
            icf = query.get("ignore_common_filters")
            if icf:
                ignore_common_filters = icf
        if q is None and query is not None:
            if isinstance(query, MetaModel):
                q = self._adapter.id_query(query.table)
            else:
                q = query
        return Set(
            self, q, ignore_common_filters=ignore_common_filters, model=model)

    def with_connection(self, f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            with self.connection():
                f(*args, **kwargs)
        return wrapped

    def atomic(self):
        return _atomic(self._adapter)

    def transaction(self):
        return _transaction(self._adapter)

    def savepoint(self):
        return _savepoint(self._adapter)

    def commit(self):
        txn = self._adapter.top_transaction()
        if txn:
            txn.commit()

    def rollback(self):
        txn = self._adapter.top_transaction()
        if txn:
            txn.rollback()


def _Database_unpickler(db_uid):
    fake_app_obj = sdict(config=sdict(db=sdict()))
    fake_app_obj.config.db.adapter = '<zombie>'
    return Database(fake_app_obj, db_uid=db_uid)


def _Database_pickler(db):
    return _Database_unpickler, (db._db_uid,)


copyreg.pickle(Database, _Database_pickler, _Database_unpickler)



# -*- coding: utf-8 -*-
"""
    emmett.language.translator
    --------------------------

    Severus translator implementation for Emmett.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

from typing import Optional
from severus.ctx import set_context
from severus.translator import Translator as _Translator

from ..ctx import current


class Translator(_Translator):
    __slots__ = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        set_context(self)

    def _update_config(self, default_language: str):
        self._default_language = default_language
        self._langmap.clear()
        self._languages.clear()
        self._build_languages()

    def _get_best_language(self, lang: Optional[str] = None) -> str:
        return self._langmap.get(
            lang or current.language, self._default_language
        )



# -*- coding: utf-8 -*-
"""
    tests.routing
    -------------

    Test Emmett routing module
"""

import pendulum
import pytest

from contextlib import contextmanager

from helpers import FakeRequestContext
from emmett import App, abort, url
from emmett.ctx import current
from emmett.http import HTTP
from emmett.testing.env import ScopeBuilder


@contextmanager
def current_ctx(app, path):
    builder = ScopeBuilder(path)
    token = current._init_(FakeRequestContext(app, builder.get_data()[0]))
    yield current
    current._close_(token)


@pytest.fixture(scope='module')
def app():
    app = App(__name__)
    app.languages = ['en', 'it']
    app.language_default = 'en'
    app.language_force_on_url = True

    @app.route()
    def test_route():
        return 'Test Router'

    @app.route()
    def test_404():
        abort(404, 'Not found, dude')

    @app.route('/test2/<int:a>/<str:b>')
    def test_route2(a, b):
        return 'Test Router'

    @app.route('/test3/<int:a>/foo(/<str:b>)?(.<str:c>)?')
    def test_route3(a, b, c):
        return 'Test Router'

    @app.route('/test_int/<int:a>')
    def test_route_int(a):
        return 'Test Router'

    @app.route('/test_float/<float:a>')
    def test_route_float(a):
        return 'Test Router'

    @app.route('/test_date/<date:a>')
    def test_route_date(a):
        return 'Test Router'

    @app.route('/test_alpha/<alpha:a>')
    def test_route_alpha(a):
        return 'Test Router'

    @app.route('/test_str/<str:a>')
    def test_route_str(a):
        return 'Test Router'

    @app.route('/test_any/<any:a>')
    def test_route_any(a):
        return 'Test Router'

    @app.route(
        '/test_complex'
        '/<int:a>'
        '/<float:b>'
        '/<date:c>'
        '/<alpha:d>'
        '/<str:e>'
        '/<any:f>'
    )
    def test_route_complex(a, b, c, d, e, f):
        return 'Test Router'

    return app


def test_routing(app):
    with current_ctx(app, '/test_int') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert not route

    with current_ctx(app, '/test_int/a') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert not route

    with current_ctx(app, '/test_int/1.1') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert not route

    with current_ctx(app, '/test_int/2000-01-01') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert not route

    with current_ctx(app, '/test_int/1') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert route.name == 'test_routing.test_route_int'

    with current_ctx(app, '/test_float') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert not route

    with current_ctx(app, '/test_float/a.a') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert not route

    with current_ctx(app, '/test_float/1') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert not route

    with current_ctx(app, '/test_float/1.1') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert route.name == 'test_routing.test_route_float'

    with current_ctx(app, '/test_date') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert not route

    with current_ctx(app, '/test_date/2000-01-01') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert route.name == 'test_routing.test_route_date'

    with current_ctx(app, '/test_alpha') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert not route

    with current_ctx(app, '/test_alpha/a1') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert not route

    with current_ctx(app, '/test_alpha/a-a') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert not route

    with current_ctx(app, '/test_alpha/a') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert route.name == 'test_routing.test_route_alpha'

    with current_ctx(app, '/test_str') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert not route

    with current_ctx(app, '/test_str/a/b') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert not route

    with current_ctx(app, '/test_str/a1-') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert route.name == 'test_routing.test_route_str'

    with current_ctx(app, '/test_any') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert not route

    with current_ctx(app, '/test_any/a/b') as ctx:
        route, args = app._router_http.match(ctx.request)
        assert route.name == 'test_routing.test_route_any'


def test_route_args(app):
    with current_ctx(
        app, '/test_complex/1/1.2/2000-12-01/foo/foo1/bar/baz'
    ) as ctx:
        route, args = app._router_http.match(ctx.request)
        assert route.name == 'test_routing.test_route_complex'
        assert args['a'] == 1
        assert args['b'] == 1.2
        assert args['c'] == pendulum.datetime(2000, 12, 1)
        assert args['d'] == 'foo'
        assert args['e'] == 'foo1'
        assert args['f'] == 'bar/baz'


@pytest.mark.asyncio
async def test_routing_valid_route(app):
    with current_ctx(app, '/it/test_route') as ctx:
        response = await app._router_http.dispatch(ctx.request, ctx.response)
        assert response.status_code == ctx.response.status == 200
        assert response.body == 'Test Router'
        assert ctx.request.language == 'it'


@pytest.mark.asyncio
async def test_routing_not_found_route(app):
    with current_ctx(app, '/') as ctx:
        with pytest.raises(HTTP) as excinfo:
            await app._router_http.dispatch(ctx.request, ctx.response)
        assert excinfo.value.status_code == 404
        assert excinfo.value.body == 'Resource not found\n'


@pytest.mark.asyncio
async def test_routing_exception_route(app):
    with current_ctx(app, '/test_404') as ctx:
        with pytest.raises(HTTP) as excinfo:
            await app._router_http.dispatch(ctx.request, ctx.response)
        assert excinfo.value.status_code == 404
        assert excinfo.value.body == 'Not found, dude'


def test_static_url(app):
    link = url('static', 'file')
    assert link == '/static/file'
    app.config.static_version_urls = True
    app.config.static_version = '1.0.0'
    link = url('static', 'js/foo.js', language='it')
    assert link == '/it/static/_1.0.0/js/foo.js'


def test_module_url(app):
    with current_ctx(app, '/') as ctx:
        ctx.request.language = 'it'
        link = url('test_route')
        assert link == '/it/test_route'
        link = url('test_route2')
        assert link == '/it/test2'
        link = url('test_route2', [2])
        assert link == '/it/test2/2'
        link = url('test_route2', [2, 'foo'])
        assert link == '/it/test2/2/foo'
        link = url('test_route3')
        assert link == '/it/test3'
        link = url('test_route3', [2])
        assert link == '/it/test3/2/foo'
        link = url('test_route3', [2, 'bar'])
        assert link == '/it/test3/2/foo/bar'
        link = url('test_route3', [2, 'bar', 'json'])
        assert link == '/it/test3/2/foo/bar.json'
        link = url(
            'test_route3', [2, 'bar', 'json'], {'foo': 'bar', 'bar': 'foo'})
        lsplit = link.split('?')
        assert lsplit[0] == '/it/test3/2/foo/bar.json'
        assert lsplit[1] in ['foo=bar&bar=foo', 'bar=foo&foo=bar']


def test_global_url_prefix(app):
    app._router_http._prefix_main = '/foo'
    app._router_http._prefix_main_len = 3

    with current_ctx(app, '/') as ctx:
        app.config.static_version_urls = False
        ctx.request.language = 'en'

        link = url('test_route')
        assert link == '/foo/test_route'

        link = url('static', 'js/foo.js')
        assert link == '/foo/static/js/foo.js'

        app.config.static_version_urls = True
        app.config.static_version = '1.0.0'

        link = url('static', 'js/foo.js')
        assert link == '/foo/static/_1.0.0/js/foo.js'

        app.config.static_version_urls = False
        ctx.request.language = 'it'

        link = url('test_route')
        assert link == '/foo/it/test_route'

        link = url('static', 'js/foo.js')
        assert link == '/foo/it/static/js/foo.js'

        app.config.static_version_urls = True
        app.config.static_version = '1.0.0'

        link = url('static', 'js/foo.js')
        assert link == '/foo/it/static/_1.0.0/js/foo.js'



# -*- coding: utf-8 -*-
"""
    emmett.orm.migrations.exceptions
    --------------------------------

    Provides exceptions for migration operations.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""


class RevisionError(Exception):
    pass


class RangeNotAncestorError(RevisionError):
    def __init__(self, lower, upper):
        self.lower = lower
        self.upper = upper
        super(RangeNotAncestorError, self).__init__(
            "Revision %s is not an ancestor of revision %s" %
            (lower or "base", upper or "base")
        )


class MultipleHeads(RevisionError):
    def __init__(self, heads, argument):
        self.heads = heads
        self.argument = argument
        super(MultipleHeads, self).__init__(
            "Multiple heads are present for given argument '%s'; "
            "%s" % (argument, ", ".join(heads))
        )


class ResolutionError(RevisionError):
    def __init__(self, message, argument):
        super(ResolutionError, self).__init__(message)
        self.argument = argument



# -*- coding: utf-8 -*-
"""
    emmett.cache
    ------------

    Provides a caching system.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import asyncio
import heapq
import os
import pickle
import tempfile
import threading
import time

from collections import OrderedDict
from functools import wraps
from typing import (
    Any, Awaitable, Callable, Dict, List, Optional, Tuple, Union, overload
)

from ._shortcuts import hashlib_sha1
from .ctx import current
from .libs.portalocker import LockedFile
from .typing import T

__all__ = ['Cache']


class CacheHashMixin:
    def __init__(self):
        self.strategies = OrderedDict()

    def add_strategy(
        self,
        key: str,
        method: Callable[..., Any] = lambda data: data
    ):
        self.strategies[key] = method

    def _hash_component(self, key: str, data: Any) -> str:
        return ''.join([key, "{", repr(data), "}"])

    def _build_hash(self, data: Dict[str, Any]) -> str:
        components = []
        for key, strategy in self.strategies.items():
            components.append(self._hash_component(key, strategy(data[key])))
        return hashlib_sha1(':'.join(components)).hexdigest()

    def _build_ctx_key(self, **ctx) -> str:
        return self.key + ":" + self._build_hash(ctx)  # type: ignore  # noqa

    @staticmethod
    def dict_strategy(data: Dict[str, Any]) -> List[Tuple[str, Any]]:
        return [(key, data[key]) for key in sorted(data)]


class CacheHandler:
    def __init__(self, prefix: str = '', default_expire: int = 300):
        self._default_expire = default_expire
        self._prefix = prefix

    @staticmethod
    def _key_prefix_(method: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(method)
        def wrap(self, key: Optional[str] = None, *args, **kwargs) -> Any:
            key = self._prefix + key if key is not None else key
            return method(self, key, *args, **kwargs)
        return wrap

    @staticmethod
    def _convert_duration_(method: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(method)
        def wrap(
            self,
            key: str,
            value: Any,
            duration: Union[int, str, None] = 'default'
        ) -> Any:
            if duration is None:
                duration = 60 * 60 * 24 * 365
            if duration == "default":
                duration = self._default_expire
            now = time.time()
            return method(
                self, key, value,
                now=now,
                duration=duration,
                expiration=now + duration  # type: ignore
            )
        return wrap

    @overload
    def __call__(
        self,
        key: Optional[str] = None,
        function: None = None,
        duration: Union[int, str, None] = 'default'
    ) -> CacheDecorator:
        ...

    @overload
    def __call__(
        self,
        key: str,
        function: Optional[Callable[..., T]],
        duration: Union[int, str, None] = 'default'
    ) -> T:
        ...

    def __call__(
        self,
        key: Optional[str] = None,
        function: Optional[Callable[..., T]] = None,
        duration: Union[int, str, None] = 'default'
    ) -> Union[CacheDecorator, T]:
        if function:
            if asyncio.iscoroutinefunction(function):
                return self.get_or_set_loop(key, function, duration)  # type: ignore
            return self.get_or_set(key, function, duration)  # type: ignore
        return CacheDecorator(self, key, duration)

    def get_or_set(
        self,
        key: str,
        function: Callable[[], T],
        duration: Union[int, str, None] = 'default'
    ) -> T:
        value = self.get(key)
        if value is None:
            value = function()
            self.set(key, value, duration)
        return value

    async def get_or_set_loop(
        self,
        key: str,
        function: Callable[[], T],
        duration: Union[int, str, None] = 'default'
    ) -> T:
        value = self.get(key)
        if value is None:
            value = await function()  # type: ignore
            self.set(key, value, duration)
        return value

    def get(self, key: str) -> Any:
        return None

    def set(self, key: str, value: Any, duration: Union[int, str, None]):
        pass

    def clear(self, key: Optional[str] = None):
        pass

    def response(
        self,
        duration: Union[int, str, None] = 'default',
        query_params: bool = True,
        language: bool = True,
        hostname: bool = False,
        headers: List[str] = []
    ) -> RouteCacheRule:
        return RouteCacheRule(
            self, query_params, language, hostname, headers, duration
        )


class CacheDecorator(CacheHashMixin):
    def __init__(
        self,
        handler: CacheHandler,
        key: Optional[str],
        duration: Union[int, str, None] = 'default'
    ):
        super().__init__()
        self._cache = handler
        self.key = key
        self.duration = duration
        self.add_strategy('args')
        self.add_strategy('kwargs', self.dict_strategy)

    def _key_from_wrapped(self, f: Callable[..., Any]) -> str:
        return f.__module__ + '.' + f.__name__

    def _wrap_sync(self, f: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(f)
        def wrap(*args, **kwargs) -> Any:
            if not args and not kwargs:
                key = self.key or self._key_from_wrapped(f)
            else:
                key = self._build_ctx_key(args=args, kwargs=kwargs)
            return self._cache.get_or_set(
                key, lambda: f(*args, **kwargs), self.duration
            )
        return wrap

    def _wrap_loop(
        self,
        f: Callable[..., Awaitable[Any]]
    ) -> Callable[..., Awaitable[Any]]:
        @wraps(f)
        async def wrap(*args, **kwargs) -> Any:
            if not args and not kwargs:
                key = self.key or self._key_from_wrapped(f)
            else:
                key = self._build_ctx_key(args=args, kwargs=kwargs)
            return await self._cache.get_or_set_loop(
                key, lambda: f(*args, **kwargs), self.duration
            )
        return wrap

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        rv = (
            self._wrap_loop(f) if asyncio.iscoroutinefunction(f) else
            self._wrap_sync(f)
        )
        if not self.key:
            self.key = f.__module__ + '.' + f.__name__
        return rv


class RamElement:
    __slots__ = ('value', 'exp', 'acc')

    def __init__(self, value: Any, exp: int, acc: int):
        self.value = value
        self.exp = exp
        self.acc = acc


class RamCache(CacheHandler):
    lock = threading.RLock()

    def __init__(
        self,
        prefix: str = '',
        threshold: int = 500,
        default_expire: int = 300
    ):
        super().__init__(
            prefix=prefix, default_expire=default_expire)
        self.data: Dict[str, Any] = {}
        self._heap_exp: List[Tuple[int, str]] = []
        self._heap_acc: List[Tuple[float, str]] = []
        self._threshold = threshold

    def _prune(self, now):
        # remove expired items
        while self._heap_exp:
            exp, rk = self._heap_exp[0]
            if exp >= now:
                break
            self._heap_exp.remove((exp, rk))
            element = self.data.get(rk)
            if element and element.exp == exp:
                self._heap_acc.remove((self.data[rk].acc, rk))
                del self.data[rk]
        # remove threshold exceding elements
        while len(self.data) > self._threshold:
            rk = heapq.heappop(self._heap_acc)[1]
            element = self.data.get(rk)
            if element:
                self._heap_exp.remove((element.exp, rk))
                del self.data[rk]

    @CacheHandler._key_prefix_
    def get(self, key: str) -> Any:
        try:
            with self.lock:
                element = self.data[key]
                now = time.time()
                if element.exp < now:
                    return None
                self._heap_acc.remove((element.acc, key))
                element.acc = now
                heapq.heappush(self._heap_acc, (element.acc, key))
            val = element.value
        except KeyError:
            return None
        return val

    @CacheHandler._key_prefix_
    @CacheHandler._convert_duration_
    def set(self, key: str, value: Any, **kwargs):
        with self.lock:
            self._prune(kwargs['now'])
            heapq.heappush(self._heap_exp, (kwargs['expiration'], key))
            heapq.heappush(self._heap_acc, (kwargs['now'], key))
            self.data[key] = RamElement(
                value, kwargs['expiration'], kwargs['now'])

    @CacheHandler._key_prefix_
    def clear(self, key: Optional[str] = None):
        with self.lock:
            if key is not None:
                try:
                    rv = self.data[key]
                    self._heap_acc.remove((rv.acc, key))
                    self._heap_exp.remove((rv.exp, key))
                    del self.data[key]
                    return
                except Exception:
                    return
            self.data.clear()
            self._heap_acc = []
            self._heap_exp = []


class DiskCache(CacheHandler):
    lock = threading.RLock()
    _fs_transaction_suffix = '.__mt_cache'
    _fs_mode = 0o600

    def __init__(
        self,
        cache_dir: str = 'cache',
        threshold: int = 500,
        default_expire: int = 300
    ):
        super().__init__(default_expire=default_expire)
        self._threshold = threshold
        self._path = os.path.join(current.app.root_path, cache_dir)
        #: create required paths if needed
        if not os.path.exists(self._path):
            os.mkdir(self._path)

    def _get_filename(self, key: str) -> str:
        khash = hashlib_sha1(key).hexdigest()
        return os.path.join(self._path, khash)

    def _del_file(self, filename: str):
        try:
            os.remove(filename)
        except Exception:
            pass

    def _list_dir(self) -> List[str]:
        return [
            os.path.join(self._path, fn)
            for fn in os.listdir(self._path)
            if not fn.endswith(self._fs_transaction_suffix)
        ]

    def _prune(self):
        with self.lock:
            entries = self._list_dir()
            if len(entries) > self._threshold:
                now = time.time()
                try:
                    for i, fpath in enumerate(entries):
                        remove = False
                        f = LockedFile(fpath, 'rb')
                        exp = pickle.load(f.file)
                        f.close()
                        remove = exp <= now or i % 3 == 0
                        if remove:
                            self._del_file(fpath)
                except Exception:
                    pass

    def get(self, key: str) -> Any:
        filename = self._get_filename(key)
        try:
            with self.lock:
                now = time.time()
                f = LockedFile(filename, 'rb')
                exp = pickle.load(f.file)
                if exp < now:
                    f.close()
                    return None
                val = pickle.load(f.file)
                f.close()
        except Exception:
            return None
        return val

    @CacheHandler._convert_duration_
    def set(self, key: str, value: Any, **kwargs):
        filename = self._get_filename(key)
        filepath = os.path.join(self._path, filename)
        with self.lock:
            self._prune()
            if os.path.exists(filepath):
                self._del_file(filepath)
            try:
                fd, tmp = tempfile.mkstemp(
                    suffix=self._fs_transaction_suffix, dir=self._path)
                with os.fdopen(fd, 'wb') as f:
                    pickle.dump(kwargs['expiration'], f, 1)
                    pickle.dump(value, f, pickle.HIGHEST_PROTOCOL)
                os.rename(tmp, filename)
                os.chmod(filename, self._fs_mode)
            except Exception:
                pass

    def clear(self, key: Optional[str] = None):
        with self.lock:
            if key is not None:
                filename = self._get_filename(key)
                try:
                    os.remove(filename)
                    return
                except Exception:
                    return
            for name in self._list_dir():
                self._del_file(name)


class RedisCache(CacheHandler):
    def __init__(
        self,
        host: str = 'localhost',
        port: int = 6379,
        password: Optional[str] = None,
        db: int = 0,
        prefix: str = 'cache:',
        default_expire: int = 300,
        **kwargs
    ):
        super().__init__(
            prefix=prefix, default_expire=default_expire)
        try:
            import redis
        except ImportError:
            raise RuntimeError('no redis module found')
        self._cache = redis.Redis(
            host=host, port=port, password=password, db=db, **kwargs
        )

    def _dump_obj(self, value: Any) -> bytes:
        if isinstance(value, int):
            return str(value).encode('ascii')
        return b'!' + pickle.dumps(value)

    def _load_obj(self, value: Any) -> Any:
        if value is None:
            return None
        if value.startswith(b'!'):
            try:
                return pickle.loads(value[1:])
            except pickle.PickleError:
                return None
        try:
            return int(value)
        except ValueError:
            return None

    @CacheHandler._key_prefix_
    def get(self, key: str) -> Any:
        return self._load_obj(self._cache.get(key))

    @CacheHandler._key_prefix_
    @CacheHandler._convert_duration_
    def set(self, key: str, value: Any, **kwargs):
        dumped = self._dump_obj(value)
        return self._cache.setex(
            name=key,
            time=kwargs['duration'],
            value=dumped
        )

    @CacheHandler._key_prefix_
    def clear(self, key: Optional[str] = None):
        if key is not None:
            if key.endswith('*'):
                keys = self._cache.delete(self._cache.keys(key))
                if keys:
                    self._cache.delete(*keys)
                return
            self._cache.delete(key)
            return
        if self._prefix:
            keys = self._cache.keys(self._prefix + '*')
            if keys:
                self._cache.delete(*keys)
            return
        self._cache.flushdb()


class RouteCacheRule(CacheHashMixin):
    def __init__(
        self,
        handler: CacheHandler,
        query_params: bool = True,
        language: bool = True,
        hostname: bool = False,
        headers: List[str] = [],
        duration: Union[int, str, None] = 'default'
    ):
        super().__init__()
        self.cache = handler
        self.check_headers = headers
        self.duration = duration
        self.add_strategy('kwargs', self.dict_strategy)
        self._ctx_builders = []
        if hostname:
            self.add_strategy('hostname')
            self._ctx_builders.append(
                ('hostname', lambda route, current: route.hostname))
        if language:
            self.add_strategy('language')
            self._ctx_builders.append(
                ('language', lambda route, current: current.language))
        if query_params:
            self.add_strategy('query_params', self.dict_strategy)
            self._ctx_builders.append(
                ('query_params', lambda route, current:
                    current.request.query_params))
        if headers:
            self.add_strategy('headers', self.headers_strategy)
            self._ctx_builders.append(
                ('headers', lambda route, current: current.request.headers))

    def _build_ctx_key(self, route: Any, **ctx) -> str:  # type: ignore
        return route.name + ":" + self._build_hash(ctx)

    def _build_ctx(
        self,
        kwargs: Dict[str, Any],
        route: Any,
        current: Any
    ) -> Dict[str, Any]:
        rv = {'kwargs': kwargs}
        for key, builder in self._ctx_builders:
            rv[key] = builder(route, current)
        return rv

    def headers_strategy(self, data: Dict[str, str]) -> List[str]:
        return [data[key] for key in self.check_headers]

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        from .routing.router import Router
        obj = Router.exposing()
        obj.cache_rule = self
        return f


class Cache:
    def __init__(self, **kwargs):
        #: load handlers
        handlers = []
        for key, val in kwargs.items():
            if key == "default":
                continue
            handlers.append((key, val))
        if not handlers:
            handlers.append(('ram', RamCache()))
        #: set handlers
        for name, handler in handlers:
            setattr(self, name, handler)
        _default_handler_name = kwargs.get('default', handlers[0][0])
        self._default_handler = getattr(self, _default_handler_name)

    @overload
    def __call__(
        self,
        key: Optional[str] = None,
        function: None = None,
        duration: Union[int, str, None] = 'default'
    ) -> CacheDecorator:
        ...

    @overload
    def __call__(
        self,
        key: str,
        function: Optional[Callable[..., T]],
        duration: Union[int, str, None] = 'default'
    ) -> T:
        ...

    def __call__(
        self,
        key: Optional[str] = None,
        function: Optional[Callable[..., T]] = None,
        duration: Union[int, str, None] = 'default'
    ) -> Union[CacheDecorator, T]:
        return self._default_handler(key, function, duration)

    def get(self, key: str) -> Any:
        return self._default_handler.get(key)

    def set(
        self,
        key: str,
        value: Any,
        duration: Union[int, str, None] = 'default'
    ):
        self._default_handler.set(key, value, duration)

    def get_or_set(
        self,
        key: str,
        function: Callable[..., T],
        duration: Union[int, str, None] = 'default'
    ) -> T:
        return self._default_handler.get_or_set(key, function, duration)

    def clear(self, key: Optional[str] = None):
        self._default_handler.clear(key)

    def response(
        self,
        duration: Union[int, str, None] = 'default',
        query_params: bool = True,
        language: bool = True,
        hostname: bool = False,
        headers: List[str] = []
    ) -> RouteCacheRule:
        return self._default_handler.response(
            duration, query_params, language, hostname, headers
        )



# -*- coding: utf-8 -*-
"""
    emmett.asgi.protocols.http.auto
    -------------------------------

    Provides HTTP auto protocol loader

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from . import protocols


if "httptools" in protocols:
    protocols.register("auto")(protocols.get("httptools"))
else:
    protocols.register("auto")(protocols.get("h11"))



# -*- coding: utf-8 -*-
"""
    emmett.asgi.loops.asyncio
    -------------------------

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import asyncio

from . import loops


@loops.register('asyncio')
def build_asyncio_loop():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    return loop



# -*- coding: utf-8 -*-
"""
    emmett.pipeline
    ---------------

    Provides the pipeline classes.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import asyncio
import types

from functools import wraps
from typing import Optional

from .helpers import flash
from .http import HTTPResponse, redirect


class Pipeline:
    __slots__ = ['_method_open', '_method_close', 'pipes']
    _type_suffix = ''

    def __init__(self, pipes=[]):
        self._method_open = f'open_{self._type_suffix}'
        self._method_close = f'close_{self._type_suffix}'
        self.pipes = pipes

    @staticmethod
    def _awaitable_wrap(f):
        @wraps(f)
        async def awaitable(*args, **kwargs):
            return f(*args, **kwargs)
        return awaitable

    def __call__(self, f):
        raise NotImplementedError

    def _flow_open(self):
        rv = []
        for pipe in self.pipes:
            if pipe._pipeline_all_methods_.issuperset(
                {'open', self._method_open}
            ):
                raise RuntimeError(
                    f'{pipe.__class__.__name__} pipe has double open methods.'
                    f' Use `open` or `{self._method_open}`, not both.'
                )
            if 'open' in pipe._pipeline_all_methods_:
                rv.append(pipe.open)
            if self._method_open in pipe._pipeline_all_methods_:
                rv.append(getattr(pipe, self._method_open))
        return rv

    def _flow_close(self):
        rv = []
        for pipe in reversed(self.pipes):
            if pipe._pipeline_all_methods_.issuperset(
                {'close', self._method_close}
            ):
                raise RuntimeError(
                    f'{pipe.__class__.__name__} pipe has double close methods.'
                    f' Use `close` or `{self._method_close}`, not both.'
                )
            if 'close' in pipe._pipeline_all_methods_:
                rv.append(pipe.close)
            if self._method_close in pipe._pipeline_all_methods_:
                rv.append(getattr(pipe, self._method_close))
        return rv


class RequestPipeline(Pipeline):
    __slots__ = []
    _type_suffix = 'request'

    def _get_proper_wrapper(self, pipe):
        if pipe._pipeline_all_methods_.issuperset(
            {'on_pipe_success', 'on_pipe_failure'}
        ):
            rv = _wrap_flow_request_complete
        elif 'on_pipe_success' in pipe._pipeline_all_methods_:
            rv = _wrap_flow_request_success
        elif 'on_pipe_failure' in pipe._pipeline_all_methods_:
            rv = _wrap_flow_request_failure
        else:
            rv = _wrap_flow_request_basic
        return rv

    def __call__(self, f):
        if not asyncio.iscoroutinefunction(f):
            f = self._awaitable_wrap(f)
        for pipe in reversed(self.pipes):
            if not isinstance(pipe, Pipe):
                continue
            if not pipe._is_flow_request_responsible:
                continue
            wrapper = self._get_proper_wrapper(pipe)
            pipe_method = (
                pipe.pipe_request
                if 'pipe_request' in pipe._pipeline_all_methods_
                else pipe.pipe)
            f = wrapper(
                pipe_method, pipe.on_pipe_success, pipe.on_pipe_failure, f)
        return f

    def _output_type(self):
        rv = None
        for pipe in reversed(self.pipes):
            if not pipe._is_flow_request_responsible or pipe.output is None:
                continue
            rv = pipe.output
        return rv


class WebsocketPipeline(Pipeline):
    __slots__ = []
    _type_suffix = 'ws'

    def _get_proper_wrapper(self, pipe):
        if pipe._pipeline_all_methods_.issuperset(
            {'on_pipe_success', 'on_pipe_failure'}
        ):
            rv = _wrap_flow_ws_complete
        elif 'on_pipe_success' in pipe._pipeline_all_methods_:
            rv = _wrap_flow_ws_success
        elif 'on_pipe_failure' in pipe._pipeline_all_methods_:
            rv = _wrap_flow_ws_failure
        else:
            rv = _wrap_flow_ws_basic
        return rv

    def __call__(self, f):
        if not asyncio.iscoroutinefunction(f):
            f = self._awaitable_wrap(f)
        for pipe in reversed(self.pipes):
            if not isinstance(pipe, Pipe):
                continue
            if not pipe._is_flow_ws_responsible:
                continue
            wrapper = self._get_proper_wrapper(pipe)
            pipe_method = (
                pipe.pipe_ws
                if 'pipe_ws' in pipe._pipeline_all_methods_
                else pipe.pipe)
            f = wrapper(
                pipe_method, pipe.on_pipe_success, pipe.on_pipe_failure, f)
        return f

    def _flow_receive(self):
        rv = []
        for pipe in self.pipes:
            if 'on_receive' not in pipe._pipeline_all_methods_:
                continue
            rv.append(pipe.on_receive)
        return rv

    def _flow_send(self):
        rv = []
        for pipe in reversed(self.pipes):
            if 'on_send' not in pipe._pipeline_all_methods_:
                continue
            rv.append(pipe.on_send)
        return rv


class MetaPipe(type):
    _pipeline_methods_ = {
        'open', 'open_request', 'open_ws',
        'close', 'close_request', 'close_ws',
        'pipe', 'pipe_request', 'pipe_ws',
        'on_pipe_success', 'on_pipe_failure',
        'on_receive', 'on_send'
    }

    def __new__(cls, name, bases, attrs):
        new_class = type.__new__(cls, name, bases, attrs)
        if not bases:
            return new_class
        declared_methods = cls._pipeline_methods_ & set(attrs.keys())
        new_class._pipeline_declared_methods_ = declared_methods
        all_methods = set()
        for base in reversed(new_class.__mro__[:-2]):
            if hasattr(base, '_pipeline_declared_methods_'):
                all_methods = all_methods | base._pipeline_declared_methods_
        all_methods = all_methods | declared_methods
        new_class._pipeline_all_methods_ = all_methods
        new_class._is_flow_request_responsible = bool(
            all_methods & {
                'pipe', 'pipe_request', 'on_pipe_success', 'on_pipe_failure'
            }
        )
        new_class._is_flow_ws_responsible = bool(
            all_methods & {
                'pipe', 'pipe_ws', 'on_pipe_success', 'on_pipe_failure'
            }
        )
        if all_methods.issuperset({'pipe', 'pipe_request'}):
            raise RuntimeError(
                f'{name} has double pipe methods. '
                'Use `pipe` or `pipe_request`, not both.'
            )
        if all_methods.issuperset({'pipe', 'pipe_ws'}):
            raise RuntimeError(
                f'{name} has double pipe methods. '
                'Use `pipe` or `pipe_ws`, not both.'
            )
        return new_class


class Pipe(metaclass=MetaPipe):
    output: Optional[str] = None

    async def open(self):
        pass

    async def open_request(self):
        pass

    async def open_ws(self):
        pass

    async def close(self):
        pass

    async def close_request(self):
        pass

    async def close_ws(self):
        pass

    async def pipe(self, next_pipe, **kwargs):
        return await next_pipe(**kwargs)

    async def pipe_request(self, next_pipe, **kwargs):
        return await next_pipe(**kwargs)

    async def pipe_ws(self, next_pipe, **kwargs):
        return await next_pipe(**kwargs)

    async def on_pipe_success(self):
        pass

    async def on_pipe_failure(self):
        pass

    def on_receive(self, data):
        return data

    def on_send(self, data):
        return data


class RequirePipe(Pipe):
    def __init__(self, condition=None, otherwise=None):
        if condition is None or otherwise is None:
            raise SyntaxError('usage: @requires(condition, otherwise)')
        if not callable(otherwise) and not isinstance(otherwise, str):
            raise SyntaxError("'otherwise' param must be string or callable")
        self.condition = condition
        self.otherwise = otherwise

    async def pipe_request(self, next_pipe, **kwargs):
        flag = self.condition()
        if not flag:
            if self.otherwise is not None:
                if callable(self.otherwise):
                    return self.otherwise()
                redirect(self.otherwise)
            else:
                flash('Insufficient privileges')
                redirect('/')
        return await next_pipe(**kwargs)

    async def pipe_ws(self, next_pipe, **kwargs):
        flag = self.condition()
        if not flag:
            return
        await next_pipe(**kwargs)


class Injector(Pipe):
    namespace: str = '__global__'

    def __init__(self):
        self._injections_ = {}
        if self.namespace != '__global__':
            self._inject = self._inject_local
            return
        self._inject = self._inject_global
        for attr_name in (
            set(dir(self)) -
            self.__class__._pipeline_methods_ -
            {'output', 'namespace'}
        ):
            if attr_name.startswith('_'):
                continue
            attr = getattr(self, attr_name)
            if isinstance(attr, types.MethodType):
                self._injections_[attr_name] = self._wrapped_method(attr)
                continue
            self._injections_[attr_name] = attr

    @staticmethod
    def _wrapped_method(method):
        def wrap(*args, **kwargs):
            return method(*args, **kwargs)
        return wrap

    def _inject_local(self, ctx):
        ctx[self.namespace] = self

    def _inject_global(self, ctx):
        ctx.update(self._injections_)

    async def pipe_request(self, next_pipe, **kwargs):
        ctx = await next_pipe(**kwargs)
        if isinstance(ctx, dict):
            self._inject(ctx)
        return ctx


def _wrap_flow_request_complete(pipe_method, on_success, on_failure, f):
    @wraps(f)
    async def flow(**kwargs):
        try:
            output = await pipe_method(f, **kwargs)
            await on_success()
            return output
        except HTTPResponse:
            await on_success()
            raise
        except Exception:
            await on_failure()
            raise
    return flow


def _wrap_flow_request_success(pipe_method, on_success, on_failure, f):
    @wraps(f)
    async def flow(**kwargs):
        try:
            output = await pipe_method(f, **kwargs)
            await on_success()
            return output
        except HTTPResponse:
            await on_success()
            raise
    return flow


def _wrap_flow_request_failure(pipe_method, on_success, on_failure, f):
    @wraps(f)
    async def flow(**kwargs):
        try:
            return await pipe_method(f, **kwargs)
        except HTTPResponse:
            raise
        except Exception:
            await on_failure()
            raise
    return flow


def _wrap_flow_request_basic(pipe_method, on_success, on_failure, f):
    @wraps(f)
    async def flow(**kwargs):
        return await pipe_method(f, **kwargs)
    return flow


def _wrap_flow_ws_complete(pipe_method, on_success, on_failure, f):
    @wraps(f)
    async def flow(**kwargs):
        try:
            await pipe_method(f, **kwargs)
            await on_success()
        except Exception:
            await on_failure()
            raise
    return flow


def _wrap_flow_ws_success(pipe_method, on_success, on_failure, f):
    @wraps(f)
    async def flow(**kwargs):
        await pipe_method(f, **kwargs)
        await on_success()
    return flow


def _wrap_flow_ws_failure(pipe_method, on_success, on_failure, f):
    @wraps(f)
    async def flow(**kwargs):
        try:
            await pipe_method(f, **kwargs)
        except Exception:
            await on_failure()
            raise
    return flow


def _wrap_flow_ws_basic(pipe_method, on_success, on_failure, f):
    @wraps(f)
    async def flow(**kwargs):
        return await pipe_method(f, **kwargs)
    return flow



# -*- coding: utf-8 -*-
"""
    emmett.orm.models
    -----------------

    Provides model layer for Emmet's ORM.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import types

from collections import OrderedDict

from ..datastructures import sdict
from ..utils import cachedprop
from .apis import (
    compute, rowattr, rowmethod, scope, belongs_to, refers_to, has_one,
    has_many
)
from .helpers import (
    Callback, ReferenceData, make_tablename, camelize, decamelize,
    wrap_scope_on_model, wrap_virtual_on_model
)
from .objects import Field, Row
from .wrappers import HasOneWrap, HasManyWrap, HasManyViaWrap


class MetaModel(type):
    _inheritable_dict_attrs_ = [
        'indexes', 'validation', ('fields_rw', {'id': False}),
        'default_values', 'update_values', 'repr_values',
        'form_labels', 'form_info', 'form_widgets'
    ]

    def __new__(cls, name, bases, attrs):
        new_class = type.__new__(cls, name, bases, attrs)
        #: collect declared attributes
        tablename = attrs.get('tablename')
        fields = []
        vfields = []
        computations = []
        callbacks = []
        declared_fields = OrderedDict()
        declared_vfields = OrderedDict()
        declared_computations = OrderedDict()
        declared_callbacks = OrderedDict()
        declared_scopes = {}
        for key, value in list(attrs.items()):
            if isinstance(value, Field):
                fields.append((key, value))
            elif isinstance(value, rowattr):
                vfields.append((key, value))
            elif isinstance(value, compute):
                computations.append((key, value))
            elif isinstance(value, Callback):
                callbacks.append((key, value))
            elif isinstance(value, scope):
                declared_scopes[key] = value
        declared_relations = sdict(
            belongs=OrderedDict(), refers=OrderedDict(),
            hasone=OrderedDict(), hasmany=OrderedDict()
        )
        for ref in belongs_to._references_.values():
            for item in ref.reference:
                rkey = list(item)[0] if isinstance(item, dict) else item
                declared_relations.belongs[rkey] = item
        belongs_to._references_ = {}
        for ref in refers_to._references_.values():
            for item in ref.reference:
                rkey = list(item)[0] if isinstance(item, dict) else item
                declared_relations.refers[rkey] = item
        refers_to._references_ = {}
        for ref in has_one._references_.values():
            for item in ref.reference:
                rkey = list(item)[0] if isinstance(item, dict) else item
                declared_relations.hasone[rkey] = item
        has_one._references_ = {}
        for ref in has_many._references_.values():
            for item in ref.reference:
                rkey = list(item)[0] if isinstance(item, dict) else item
                declared_relations.hasmany[rkey] = item
        has_many._references_ = {}
        #: sort declared attributes that keeps order
        fields.sort(key=lambda x: x[1]._inst_count_)
        vfields.sort(key=lambda x: x[1]._inst_count_)
        computations.sort(key=lambda x: x[1]._inst_count_)
        callbacks.sort(key=lambda x: x[1]._inst_count_)
        declared_fields.update(fields)
        declared_vfields.update(vfields)
        declared_computations.update(computations)
        declared_callbacks.update(callbacks)
        #: store declared attributes in class
        new_class._declared_tablename_ = tablename
        new_class._declared_fields_ = declared_fields
        new_class._declared_virtuals_ = declared_vfields
        new_class._declared_computations_ = declared_computations
        new_class._declared_callbacks_ = declared_callbacks
        new_class._declared_scopes_ = declared_scopes
        new_class._declared_belongs_ref_ = declared_relations.belongs
        new_class._declared_refers_ref_ = declared_relations.refers
        new_class._declared_hasone_ref_ = declared_relations.hasone
        new_class._declared_hasmany_ref_ = declared_relations.hasmany
        #: get super declared attributes
        all_fields = OrderedDict()
        all_vfields = OrderedDict()
        all_computations = OrderedDict()
        all_callbacks = OrderedDict()
        all_scopes = {}
        all_relations = sdict(
            belongs=OrderedDict(), refers=OrderedDict(),
            hasone=OrderedDict(), hasmany=OrderedDict()
        )
        for base in reversed(new_class.__mro__[1:]):
            if hasattr(base, '_declared_fields_'):
                all_fields.update(base._declared_fields_)
            if hasattr(base, '_declared_virtuals_'):
                all_vfields.update(base._declared_virtuals_)
            if hasattr(base, '_declared_computations_'):
                all_computations.update(base._declared_computations_)
            if hasattr(base, '_declared_callbacks_'):
                all_callbacks.update(base._declared_callbacks_)
            if hasattr(base, '_declared_scopes_'):
                all_scopes.update(base._declared_scopes_)
            for key in list(all_relations):
                attrkey = '_declared_' + key + '_ref_'
                if hasattr(base, attrkey):
                    all_relations[key].update(getattr(base, attrkey))
        #: compose 'all' attributes
        all_fields.update(declared_fields)
        all_vfields.update(declared_vfields)
        all_computations.update(declared_computations)
        all_callbacks.update(declared_callbacks)
        all_scopes.update(declared_scopes)
        for key in list(all_relations):
            all_relations[key].update(declared_relations[key])
        #: store 'all' attributes on class
        new_class._all_fields_ = all_fields
        new_class._all_virtuals_ = all_vfields
        new_class._all_computations_ = all_computations
        new_class._all_callbacks_ = all_callbacks
        new_class._all_scopes_ = all_scopes
        new_class._all_belongs_ref_ = all_relations.belongs
        new_class._all_refers_ref_ = all_relations.refers
        new_class._all_hasone_ref_ = all_relations.hasone
        new_class._all_hasmany_ref_ = all_relations.hasmany
        return new_class


class Model(metaclass=MetaModel):
    db = None
    table = None

    auto_validation = True

    @classmethod
    def _init_inheritable_dicts_(cls):
        if cls.__bases__ != (object,):
            return
        for attr in cls._inheritable_dict_attrs_:
            if isinstance(attr, tuple):
                attr_name, default = attr
            else:
                attr_name, default = attr, {}
            if not isinstance(default, dict):
                raise SyntaxError(f"{attr_name} is not a dictionary")
            setattr(cls, attr_name, default)

    @classmethod
    def _merge_inheritable_dicts_(cls, models):
        for attr in cls._inheritable_dict_attrs_:
            if isinstance(attr, tuple):
                attr_name = attr[0]
            else:
                attr_name = attr
            attrs = {}
            for model in models:
                superattrs = getattr(model, attr_name)
                for k, v in superattrs.items():
                    attrs[k] = v
            for k, v in getattr(cls, attr_name).items():
                attrs[k] = v
            setattr(cls, attr_name, attrs)

    @classmethod
    def __getsuperattrs(cls):
        superattr = "_supermodels" + cls.__name__
        if hasattr(cls, superattr):
            return
        supermodels = cls.__bases__
        superattr_val = []
        for supermodel in supermodels:
            try:
                supermodel.__getsuperattrs()
                superattr_val.append(supermodel)
            except Exception:
                pass
        setattr(cls, superattr, superattr_val)
        sup = getattr(cls, superattr)
        if not sup:
            return
        cls._merge_inheritable_dicts_(sup)

    def __new__(cls):
        if cls._declared_tablename_ is None:
            cls.tablename = make_tablename(cls.__name__)
        cls.__getsuperattrs()
        return super(Model, cls).__new__(cls)

    def __init__(self):
        if not hasattr(self, 'migrate'):
            self.migrate = self.config.get('migrate', self.db._migrate)
        if not hasattr(self, 'format'):
            self.format = None

    @property
    def config(self):
        return self.db.config

    def __parse_relation_via(self, via):
        if via is None:
            return via
        rv = sdict()
        splitted = via.split('.')
        rv.via = splitted[0]
        if len(splitted) > 1:
            rv.field = splitted[1]
        return rv

    def __parse_belongs_relation(self, item, on_delete):
        rv = sdict(on_delete=on_delete)
        if isinstance(item, dict):
            rv.name = list(item)[0]
            rdata = item[rv.name]
            if isinstance(rdata, dict):
                if "target" in rdata:
                    rv.model = rdata["target"]
                else:
                    rv.model = camelize(rv.name)
                if "on_delete" in rdata:
                    rv.on_delete = rdata["on_delete"]
            else:
                rv.model = rdata
            if rv.model == "self":
                rv.model = self.__class__.__name__
        else:
            rv.name = item
            rv.model = camelize(item)
        return rv

    def __build_relation_modelname(self, name, relation, singularize):
        relation.model = camelize(name)
        if singularize:
            relation.model = relation.model[:-1]

    def __build_relation_fieldname(self, relation):
        splitted = relation.model.split('.')
        relation.model = splitted[0]
        if len(splitted) > 1:
            relation.field = splitted[1]
        else:
            relation.field = decamelize(self.__class__.__name__)

    def __parse_relation_dict(self, rel, singularize):
        if 'scope' in rel.model:
            rel.scope = rel.model['scope']
        if 'where' in rel.model:
            rel.where = rel.model['where']
        if 'via' in rel.model:
            rel.update(self.__parse_relation_via(rel.model['via']))
            del rel.model
        else:
            if 'target' in rel.model:
                rel.model = rel.model['target']
            if not isinstance(rel.model, str):
                self.__build_relation_modelname(rel.name, rel, singularize)

    def __parse_many_relation(self, item, singularize=True):
        rv = ReferenceData(self)
        if isinstance(item, dict):
            rv.name = list(item)[0]
            rv.model = item[rv.name]
            if isinstance(rv.model, dict):
                if 'method' in rv.model:
                    rv.field = rv.model.get(
                        'field', decamelize(self.__class__.__name__)
                    )
                    rv.cast = rv.model.get('cast')
                    rv.method = rv.model['method']
                    del rv.model
                else:
                    self.__parse_relation_dict(rv, singularize)
        else:
            rv.name = item
            self.__build_relation_modelname(item, rv, singularize)
        if rv.model:
            if not rv.field:
                self.__build_relation_fieldname(rv)
            if rv.model == "self":
                rv.model = self.__class__.__name__
        return rv

    def _define_props_(self):
        #: create pydal's Field elements
        self.fields = []
        idfield = Field('id')._make_field('id', self)
        setattr(self.__class__, 'id', idfield)
        self.fields.append(idfield)
        for name, obj in self._all_fields_.items():
            if obj.modelname is not None:
                obj = Field(obj._type, *obj._args, **obj._kwargs)
                setattr(self.__class__, name, obj)
            self.fields.append(obj._make_field(name, self))

    def _define_relations_(self):
        _ftype_builder = lambda v: 'reference {}'.format(v)
        self._virtual_relations_ = OrderedDict()
        bad_args_error = (
            "belongs_to, has_one and has_many "
            "only accept strings or dicts as arguments"
        )
        #: belongs_to and refers_to are mapped with 'reference' type Field
        _references = []
        _reference_keys = ['_all_belongs_ref_', '_all_refers_ref_']
        belongs_references = {}
        for key in _reference_keys:
            if hasattr(self, key):
                _references.append(list(getattr(self, key).values()))
            else:
                _references.append([])
        isbelongs, ondelete = True, 'cascade'
        for _references_obj in _references:
            for item in _references_obj:
                if not isinstance(item, (str, dict)):
                    raise RuntimeError(bad_args_error)
                reference = self.__parse_belongs_relation(item, ondelete)
                if reference.model != self.__class__.__name__:
                    tablename = self.db[reference.model]._tablename
                else:
                    tablename = self.tablename
                fieldobj = Field(
                    _ftype_builder(tablename),
                    ondelete=reference.on_delete,
                    _isrefers=not isbelongs
                )
                setattr(self.__class__, reference.name, fieldobj)
                self.fields.append(
                    getattr(self, reference.name)._make_field(
                        reference.name, self
                    )
                )
                belongs_references[reference.name] = reference.model
            isbelongs = False
            ondelete = 'nullify'
        setattr(self.__class__, '_belongs_ref_', belongs_references)
        #: has_one are mapped with rowattr
        hasone_references = {}
        if hasattr(self, '_all_hasone_ref_'):
            for item in getattr(self, '_all_hasone_ref_').values():
                if not isinstance(item, (str, dict)):
                    raise RuntimeError(bad_args_error)
                reference = self.__parse_many_relation(item, False)
                self._virtual_relations_[reference.name] = rowattr(
                    reference.name
                )(HasOneWrap(reference))
                hasone_references[reference.name] = reference
        setattr(self.__class__, '_hasone_ref_', hasone_references)
        #: has_many are mapped with rowattr
        hasmany_references = {}
        if hasattr(self, '_all_hasmany_ref_'):
            for item in getattr(self, '_all_hasmany_ref_').values():
                if not isinstance(item, (str, dict)):
                    raise RuntimeError(bad_args_error)
                reference = self.__parse_many_relation(item)
                if reference.via is not None:
                    #: maps has_many({'things': {'via': 'otherthings'}})
                    wrapper = HasManyViaWrap
                else:
                    #: maps has_many('things'),
                    #  has_many({'things': 'othername'})
                    wrapper = HasManyWrap
                self._virtual_relations_[reference.name] = rowattr(
                    reference.name
                )(wrapper(reference))
                hasmany_references[reference.name] = reference
        setattr(self.__class__, '_hasmany_ref_', hasmany_references)

    def _define_virtuals_(self):
        self._all_rowattrs_ = {}
        self._all_rowmethods_ = {}
        err = 'rowattr or rowmethod cannot have the name of an existent field!'
        field_names = [field.name for field in self.fields]
        for attr in ['_virtual_relations_', '_all_virtuals_']:
            for _, obj in getattr(self, attr, {}).items():
                if obj.field_name in field_names:
                    raise RuntimeError(err)
                wrapped = wrap_virtual_on_model(self, obj.f)
                if isinstance(obj, rowmethod):
                    self._all_rowmethods_[obj.field_name] = wrapped
                    f = Field.Method(obj.field_name, wrapped)
                else:
                    self._all_rowattrs_[obj.field_name] = wrapped
                    f = Field.Virtual(obj.field_name, wrapped)
                self.fields.append(f)

    def _build_rowclass_(self):
        clsname = self.__class__.__name__ + "Row"
        attrs = {
            k: cachedprop(v, name=k) for k, v in self._all_rowattrs_.items()
        }
        attrs.update(self._all_rowmethods_)
        self._rowclass_ = type(clsname, (Row,), attrs)
        globals()[clsname] = self._rowclass_

    def _define_(self):
        self.__define_indexes()
        self.__define_validation()
        self.__define_access()
        self.__define_defaults()
        self.__define_updates()
        self.__define_representation()
        self.__define_computations()
        self.__define_callbacks()
        self.__define_scopes()
        self.__define_form_utils()
        self.setup()

    def __define_validation(self):
        for field in self.fields:
            if isinstance(field, (Field.Method, Field.Virtual)):
                continue
            validation = self.validation.get(field.name, {})
            if isinstance(validation, dict):
                for key in list(validation):
                    field._requires[key] = validation[key]
            elif isinstance(validation, list):
                field._custom_requires += validation
            else:
                field._custom_requires.append(validation)
            field._parse_validation()

    def __define_access(self):
        for field, value in self.fields_rw.items():
            if isinstance(value, (tuple, list)):
                readable, writable = value
            else:
                writable = value
                readable = value
            self.table[field].writable = writable
            self.table[field].readable = readable

    def __define_defaults(self):
        for field, value in self.default_values.items():
            self.table[field].default = value

    def __define_updates(self):
        for field, value in self.update_values.items():
            self.table[field].update = value

    def __define_representation(self):
        for field, value in self.repr_values.items():
            self.table[field].represent = value

    def __define_computations(self):
        err = 'computations should have the name of an existing field to compute!'
        field_names = [field.name for field in self.fields]
        for obj in self._all_computations_.values():
            if obj.field_name not in field_names:
                raise RuntimeError(err)
            # TODO add check virtuals
            self.table[obj.field_name].compute = (
                lambda row, obj=obj, self=self: obj.f(self, row)
            )

    def __define_callbacks(self):
        for obj in self._all_callbacks_.values():
            for t in obj.t:
                if t in ["_before_insert", "_before_delete", "_after_delete"]:
                    getattr(self.table, t).append(
                        lambda a, obj=obj, self=self: obj.f(self, a)
                    )
                else:
                    getattr(self.table, t).append(
                        lambda a, b, obj=obj, self=self: obj.f(self, a, b)
                    )

    def __define_scopes(self):
        self._scopes_ = {}
        for obj in self._all_scopes_.values():
            self._scopes_[obj.name] = obj
            if not hasattr(self.__class__, obj.name):
                setattr(
                    self.__class__, obj.name,
                    classmethod(wrap_scope_on_model(obj.f))
                )

    def __prepend_table_on_index_name(self, name):
        return '%s_widx__%s' % (self.tablename, name)

    def __create_index_name(self, *values):
        components = []
        for value in values:
            components.append(value.replace('_', ''))
        return self.__prepend_table_on_index_name("_".join(components))

    def __parse_index_dict(self, value):
        rv = {}
        fields = value.get('fields') or []
        if not isinstance(fields, (list, tuple)):
            fields = [fields]
        rv['fields'] = fields
        where_query = None
        where_cond = value.get('where')
        if callable(where_cond):
            where_query = where_cond(self.__class__)
        if where_query:
            rv['where'] = where_query
        expressions = []
        expressions_cond = value.get('expressions')
        if callable(expressions_cond):
            expressions = expressions_cond(self.__class__)
        if not isinstance(expressions, (tuple, list)):
            expressions = [expressions]
        rv['expressions'] = expressions
        rv['unique'] = value.get('unique', False)
        return rv

    def __define_indexes(self):
        self._indexes_ = {}
        for key, value in self.indexes.items():
            if isinstance(value, bool):
                if not value:
                    continue
                if not isinstance(key, tuple):
                    key = [key]
                if any(field not in self.table for field in key):
                    raise SyntaxError(f'Invalid field specified in indexes: {key}')
                idx_name = self.__create_index_name(*key)
                idx_dict = {'fields': key, 'expressions': [], 'unique': False}
            elif isinstance(value, dict):
                idx_name = self.__prepend_table_on_index_name(key)
                idx_dict = self.__parse_index_dict(value)
            else:
                raise SyntaxError('Values in indexes dict should be booleans or dicts')
            self._indexes_[idx_name] = idx_dict

    def __define_form_utils(self):
        #: labels
        for field, value in self.form_labels.items():
            self.table[field].label = value
        #: info
        for field, value in self.form_info.items():
            self.table[field].comment = value
        #: widgets
        for field, value in self.form_widgets.items():
            self.table[field].widget = value

    def setup(self):
        pass

    @classmethod
    def _instance_(cls):
        return cls.table._model_

    @classmethod
    def new(cls, **attributes):
        row = cls._instance_()._rowclass_()
        for field in cls.table.fields:
            val = attributes.get(field, cls.table[field].default)
            if callable(val):
                val = val()
            row[field] = val
        return row

    @classmethod
    def create(cls, *args, **kwargs):
        if args:
            if isinstance(args[0], (dict, sdict)):
                for key in list(args[0]):
                    kwargs[key] = args[0][key]
        return cls.table.validate_and_insert(**kwargs)

    @classmethod
    def validate(cls, row):
        row = sdict(row)
        errors = sdict()
        for field in cls.table.fields:
            default = getattr(cls.table[field], 'default')
            if callable(default):
                default = default()
            value = row.get(field, default)
            _, error = cls.table[field].validate(value)
            if error:
                errors[field] = error
        return errors

    @classmethod
    def where(cls, cond):
        if not isinstance(cond, types.LambdaType):
            raise ValueError("Model.where expects a lambda as parameter")
        return cls.db.where(cond(cls), model=cls)

    @classmethod
    def all(cls):
        return cls.db.where(cls.table, model=cls)

    @classmethod
    def first(cls):
        return cls.all().select(orderby=cls.id, limitby=(0, 1)).first()

    @classmethod
    def last(cls):
        return cls.all().select(orderby=~cls.id, limitby=(0, 1)).first()

    @classmethod
    def get(cls, *args, **kwargs):
        if len(args) == 1:
            return cls.table[args[0]]
        return cls.table(**kwargs)

    @rowmethod('update_record')
    def _update_record(self, row, **fields):
        newfields = fields or dict(row)
        for fieldname in list(newfields.keys()):
            if (
                fieldname not in self.table.fields or
                self.table[fieldname].type == 'id'
            ):
                del newfields[fieldname]
        self.db(self.table._id == row.id, ignore_common_filters=True).update(
            **newfields
        )
        row.update(newfields)
        return row

    @rowmethod('delete_record')
    def _delete_record(self, row):
        return self.db(self.db[self.tablename]._id == row.id).delete()



# -*- coding: utf-8 -*-
"""
    tests.session
    -------------

    Test Emmett session module
"""

import pytest

from emmett.asgi.handlers import RequestContext
from emmett.ctx import current
from emmett.sessions import SessionManager
from emmett.testing.env import ScopeBuilder
from emmett.wrappers.request import Request
from emmett.wrappers.response import Response


class FakeRequestContext(RequestContext):
    def __init__(self, app, scope):
        self.request = Request(scope, None, None)
        self.response = Response()
        self.session = None


@pytest.fixture(scope='module')
def ctx():
    builder = ScopeBuilder()
    token = current._init_(FakeRequestContext(None, builder.get_data()[0]))
    yield current
    current._close_(token)


@pytest.mark.asyncio
async def test_session_cookie(ctx):
    session_cookie = SessionManager.cookies(
        key='sid',
        secure=True,
        domain='localhost',
        cookie_name='foo_session'
    )
    assert session_cookie.key == 'sid'
    assert session_cookie.secure is True
    assert session_cookie.domain == 'localhost'

    await session_cookie.open_request()
    assert ctx.session._expiration == 3600

    await session_cookie.close_request()
    cookie = str(ctx.response.cookies)
    assert 'foo_session' in cookie
    assert 'Domain=localhost;' in cookie
    assert 'secure' in cookie.lower()

    ctx.request.cookies = ctx.response.cookies
    await session_cookie.open_request()
    assert ctx.session._expiration == 3600



# -*- coding: utf-8 -*-
"""
    emmett.asgi.protocols.http.h2
    -----------------------------

    Provides HTTP h2 protocol implementation

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import asyncio

from collections import defaultdict
from functools import partial
from typing import Any, Callable, List, Optional, Tuple
from urllib.parse import unquote

import h11
import h2.config
import h2.connection
import h2.events
import h2.exceptions

from uvicorn.protocols.http.flow_control import HIGH_WATER_LIMIT
from uvicorn.protocols.utils import get_client_addr, get_path_with_query_string

from .helpers import (
    TRACE_LOG_LEVEL,
    ASGICycle,
    Config,
    HTTPProtocol,
    ServerState,
    _service_unavailable
)


class EventsRegistry(defaultdict):
    def __init__(self):
        super().__init__(lambda: (lambda *args, **kwargs: None))

    def register(self, key: Any):
        def wrap(f):
            self[key] = f
            return f
        return wrap


class H2Protocol(HTTPProtocol):
    __slots__ = ["conn", "streams"]

    alpn_protocols = ["h2"]

    def __init__(
        self,
        config: Config,
        server_state: ServerState,
        on_connection_lost: Optional[Callable[[], None]] = None,
        _loop=None
    ):
        super().__init__(
            config=config,
            server_state=server_state,
            on_connection_lost=on_connection_lost,
            _loop=_loop
        )
        self.conn = h2.connection.H2Connection(
            config=h2.config.H2Configuration(
                client_side=False,
                header_encoding=None
            )
        )
        self.streams = {}

    def connection_made(self, transport: asyncio.Transport, init: bool = False):
        super().connection_made(transport)
        if init:
            self.conn.initiate_connection()
            self.transport.write(self.conn.data_to_send())

    def connection_lost(self, exc):
        self.connections.discard(self)

        if self.logger.level <= TRACE_LOG_LEVEL:
            prefix = "%s:%d - " % tuple(self.addr_remote) if self.addr_remote else ""
            self.logger.log(TRACE_LOG_LEVEL, "%sConnection lost", prefix)

        for stream in self.streams.values():
            stream.message_event.set()
        if self.flow is not None:
            self.flow.resume_writing()
        if exc is None:
            self.transport.close()

        if self.on_connection_lost:
            self.on_connection_lost()

    def handle_upgrade_from_h11(
        self,
        transport: asyncio.Protocol,
        upgrade_event: h11.Request,
        headers: List[Tuple[bytes, bytes]]
    ):
        self.connection_made(transport, init=False)

        settings = ""
        headers = [
            (b":method", upgrade_event.method.encode("ascii")),
            (b":path", upgrade_event.target),
            (b":scheme", self.scheme.encode("ascii"))
        ]
        for name, value in headers:
            if name == b"http2-settings":
                settings = value.decode("latin-1")
            elif name == b"host":
                headers.append((b":authority", value))
            else:
                headers.append((name, value))

        self.conn.initiate_upgrade_connection(settings)
        self.transport.write(self.conn.data_to_send())
        event = h2.events.RequestReceived()
        event.stream_id = 1
        event.headers = headers
        on_request_received(self, event)

    def data_received(self, data: bytes):
        self._might_unset_keepalive()

        try:
            events = self.conn.receive_data(data)
        except h2.exceptions.ProtocolError:
            self.transport.write(self.conn.data_to_send())
            self.transport.close()
            return

        for event in events:
            eventsreg[type(event)](self, event)

    def shutdown(self):
        self.transport.write(self.conn.data_to_send())
        self.transport.close()

    def timeout_keep_alive_handler(self):
        if self.transport.is_closing():
            return

        self.conn.close_connection()
        self.transport.write(self.conn.data_to_send())
        self.transport.close()

    def on_response_complete(self, stream_id):
        self.server_state.total_requests += 1
        self.streams.pop(stream_id, None)
        if self.transport.is_closing():
            return

        if not self.streams:
            self._might_unset_keepalive()
            self.timeout_keep_alive_task = self.loop.call_later(
                self.timeout_keep_alive, self.timeout_keep_alive_handler
            )
        self.flow.resume_reading()


class H2ASGICycle(ASGICycle):
    __slots__ = ["scheme", "host", "stream_id", "new_request"]

    def __init__(
        self,
        scope,
        conn,
        protocol: H2Protocol,
        stream_id: int,
        host: bytes
    ):
        super().__init__(scope, conn, protocol)
        self.scheme = protocol.scheme.encode("ascii")
        self.host = host
        self.stream_id = stream_id
        self.new_request = partial(on_request_received, protocol)

    async def send(self, message):
        message_type = message["type"]

        if self.flow.write_paused and not self.disconnected:
            await self.flow.drain()

        if self.disconnected:
            return

        if message_type == "http.response.push":
            push_stream_id = self.conn.get_next_available_stream_id()
            headers = [
                (b":authority", self.host),
                (b":method", b"GET"),
                (b":path", message["path"].encode("ascii")),
                (b":scheme", self.scheme)
            ] + message["headers"]

            try:
                self.conn.push_stream(
                    stream_id=self.stream_id,
                    promised_stream_id=push_stream_id,
                    request_headers=headers
                )
                self.transport.write(self.conn.data_to_send())
            except h2.exceptions.ProtocolError:
                self.logger.debug("h2 protocol error.", exc_info=True)
            else:
                event = h2.events.RequestReceived()
                event.stream_id = push_stream_id
                event.headers = headers
                self.new_request(event)
        elif not self.response_started:
            # Sending response status line and headers
            if message_type != "http.response.start":
                msg = "Expected ASGI message 'http.response.start', but got '%s'."
                raise RuntimeError(msg % message_type)

            self.response_started = True

            status_code = message["status"]
            headers = (
                [(":status", str(status_code))] +
                self.default_headers +
                message.get("headers", [])
            )

            if self.access_log_enabled:
                self.access_logger.info(
                    '%s - "%s %s HTTP/%s" %d',
                    get_client_addr(self.scope),
                    self.scope["method"],
                    get_path_with_query_string(self.scope),
                    self.scope["http_version"],
                    status_code
                )

            self.conn.send_headers(self.stream_id, headers, end_stream=False)
            self.transport.write(self.conn.data_to_send())
        elif not self.response_completed:
            if message_type == "http.response.body":
                more_body = message.get("more_body", False)
                if self.scope["method"] == "HEAD":
                    body = b""
                else:
                    body = message.get("body", b"")
                self.conn.send_data(self.stream_id, body, end_stream=not more_body)
                self.transport.write(self.conn.data_to_send())
                if not more_body:
                    self.response_completed = True
                    self.message_event.set()
            else:
                msg = "Got unexpected ASGI message '%s'."
                raise RuntimeError(msg % message_type)

        if self.response_completed:
            self.on_response(self.stream_id)


eventsreg = EventsRegistry()


@eventsreg.register(h2.events.RequestReceived)
def on_request_received(protocol: H2Protocol, event: h2.events.RequestReceived):
    headers, pseudo_headers = [], {}
    for key, value in event.headers:
        if key[0] == b":"[0]:
            pseudo_headers[key] = value
        else:
            headers.append((key.lower(), value))
    host = pseudo_headers[b":authority"]
    headers.append((b"host", host))

    raw_path, _, query_string = pseudo_headers[b":path"].partition(b"?")
    scope = {
        "type": "http",
        "asgi": {
            "version": protocol.config.asgi_version,
            "spec_version": "2.1"
        },
        "http_version": "2",
        "server": protocol.addr_local,
        "client": protocol.addr_remote,
        "scheme": protocol.scheme,
        "method": pseudo_headers[b":method"].decode("ascii"),
        "root_path": protocol.root_path,
        "path": unquote(raw_path.decode("ascii")),
        "raw_path": raw_path,
        "query_string": query_string,
        "headers": headers,
        "extensions": {"http.response.push": {}}
    }

    if protocol.limit_concurrency is not None and (
        len(protocol.connections) >= protocol.limit_concurrency or
        len(protocol.tasks) >= protocol.limit_concurrency
    ):
        app = _service_unavailable
        message = "Exceeded concurrency limit."
        protocol.logger.warning(message)
    else:
        app = protocol.app

    protocol.streams[event.stream_id] = cycle = H2ASGICycle(
        scope=scope,
        conn=protocol.conn,
        protocol=protocol,
        stream_id=event.stream_id,
        host=host
    )
    task = protocol.loop.create_task(cycle.run_asgi(app))
    task.add_done_callback(protocol.tasks.discard)
    protocol.tasks.add(task)


@eventsreg.register(h2.events.DataReceived)
def on_data_received(protocol: H2Protocol, event: h2.events.DataReceived):
    try:
        stream = protocol.streams[event.stream_id]
    except KeyError:
        protocol.conn.reset_stream(
            event.stream_id,
            error_code=h2.errors.ErrorCodes.PROTOCOL_ERROR
        )
        return

    stream.body += event.data
    if len(stream.body) > HIGH_WATER_LIMIT:
        protocol.flow.pause_reading()
    stream.message_event.set()


@eventsreg.register(h2.events.StreamEnded)
def on_stream_ended(protocol: H2Protocol, event: h2.events.StreamEnded):
    try:
        stream = protocol.streams[event.stream_id]
    except KeyError:
        protocol.conn.reset_stream(
            event.stream_id,
            error_code=h2.errors.ErrorCodes.STREAM_CLOSED
        )
        return

    stream.transport.resume_reading()
    stream.more_body = False
    stream.message_event.set()


@eventsreg.register(h2.events.ConnectionTerminated)
def on_connection_terminated(
    protocol: H2Protocol,
    event: h2.events.ConnectionTerminated
):
    stream = protocol.streams.pop(event.last_stream_id, None)
    if stream:
        stream.disconnected = True
    protocol.conn.close_connection(last_stream_id=event.last_stream_id)
    protocol.transport.write(protocol.conn.data_to_send())
    protocol.transport.close()


@eventsreg.register(h2.events.StreamReset)
def on_stream_reset(protocol: H2Protocol, event: h2.events.StreamReset):
    protocol.streams.pop(event.stream_id, None)
